import operator
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Annotated, Any, Literal, Optional, cast

import strawberry
from aioitertools.itertools import groupby, islice
from openinference.semconv.trace import SpanAttributes
from sqlalchemy import and_, case, desc, distinct, exists, func, or_, select
from sqlalchemy.dialects import postgresql, sqlite
from sqlalchemy.sql.expression import tuple_
from sqlalchemy.sql.functions import percentile_cont
from strawberry import ID, UNSET, lazy
from strawberry.relay import Connection, Edge, Node, NodeID, PageInfo
from strawberry.types import Info
from typing_extensions import assert_never

from phoenix.datetime_utils import get_timestamp_range, normalize_datetime, right_open_time_range
from phoenix.db import models
from phoenix.db.helpers import SupportedSQLDialect, date_trunc
from phoenix.server.api.context import Context
from phoenix.server.api.exceptions import BadRequest
from phoenix.server.api.input_types.ProjectSessionSort import (
    ProjectSessionSort,
    ProjectSessionSortConfig,
)
from phoenix.server.api.input_types.SpanSort import SpanColumn, SpanSort, SpanSortConfig
from phoenix.server.api.input_types.TimeBinConfig import TimeBinConfig, TimeBinScale
from phoenix.server.api.input_types.TimeRange import TimeRange
from phoenix.server.api.types.AnnotationConfig import AnnotationConfig, to_gql_annotation_config
from phoenix.server.api.types.AnnotationSummary import AnnotationSummary
from phoenix.server.api.types.CostBreakdown import CostBreakdown
from phoenix.server.api.types.DocumentEvaluationSummary import DocumentEvaluationSummary
from phoenix.server.api.types.GenerativeModel import GenerativeModel
from phoenix.server.api.types.pagination import (
    ConnectionArgs,
    Cursor,
    CursorSortColumn,
    CursorSortColumnDataType,
    CursorString,
    connection_from_cursors_and_nodes,
    connection_from_list,
)
from phoenix.server.api.types.ProjectSession import ProjectSession
from phoenix.server.api.types.SortDir import SortDir
from phoenix.server.api.types.Span import Span
from phoenix.server.api.types.SpanCostSummary import SpanCostSummary
from phoenix.server.api.types.TimeSeries import TimeSeries, TimeSeriesDataPoint
from phoenix.server.api.types.Trace import Trace
from phoenix.server.api.types.ValidationResult import ValidationResult
from phoenix.server.session_filters import get_filtered_session_rowids_subquery
from phoenix.server.types import DbSessionFactory
from phoenix.trace.dsl import SpanFilter

DEFAULT_PAGE_SIZE = 30
if TYPE_CHECKING:
    from phoenix.server.api.types.ProjectTraceRetentionPolicy import ProjectTraceRetentionPolicy


@strawberry.type
class Project(Node):
    id: NodeID[int]
    db_record: strawberry.Private[Optional[models.Project]] = None

    def __post_init__(self) -> None:
        if self.db_record and self.id != self.db_record.id:
            raise ValueError("Project ID mismatch")

    @strawberry.field
    async def name(
        self,
        info: Info[Context, None],
    ) -> str:
        if self.db_record:
            name = self.db_record.name
        else:
            name = await info.context.data_loaders.project_fields.load(
                (self.id, models.Project.name),
            )
        return name

    @strawberry.field
    async def gradient_start_color(
        self,
        info: Info[Context, None],
    ) -> str:
        if self.db_record:
            gradient_start_color = self.db_record.gradient_start_color
        else:
            gradient_start_color = await info.context.data_loaders.project_fields.load(
                (self.id, models.Project.gradient_start_color),
            )
        return gradient_start_color

    @strawberry.field
    async def gradient_end_color(
        self,
        info: Info[Context, None],
    ) -> str:
        if self.db_record:
            gradient_end_color = self.db_record.gradient_end_color
        else:
            gradient_end_color = await info.context.data_loaders.project_fields.load(
                (self.id, models.Project.gradient_end_color),
            )
        return gradient_end_color

    @strawberry.field
    async def start_time(
        self,
        info: Info[Context, None],
    ) -> Optional[datetime]:
        start_time = await info.context.data_loaders.min_start_or_max_end_times.load(
            (self.id, "start"),
        )
        start_time, _ = right_open_time_range(start_time, None)
        return start_time

    @strawberry.field
    async def end_time(
        self,
        info: Info[Context, None],
    ) -> Optional[datetime]:
        end_time = await info.context.data_loaders.min_start_or_max_end_times.load(
            (self.id, "end"),
        )
        _, end_time = right_open_time_range(None, end_time)
        return end_time

    @strawberry.field
    async def record_count(
        self,
        info: Info[Context, None],
        time_range: Optional[TimeRange] = UNSET,
        filter_condition: Optional[str] = UNSET,
        session_filter_condition: Optional[str] = UNSET,
    ) -> int:
        if filter_condition and session_filter_condition:
            raise BadRequest(
                "Both a filter condition and session filter condition "
                "cannot be applied at the same time"
            )
        return await info.context.data_loaders.record_counts.load(
            (
                "span",
                self.id,
                time_range or None,
                filter_condition or None,
                session_filter_condition or None,
            ),
        )

    @strawberry.field
    async def trace_count(
        self,
        info: Info[Context, None],
        time_range: Optional[TimeRange] = UNSET,
        filter_condition: Optional[str] = UNSET,
        session_filter_condition: Optional[str] = UNSET,
    ) -> int:
        if filter_condition and session_filter_condition:
            raise BadRequest(
                "Both a filter condition and session filter condition "
                "cannot be applied at the same time"
            )
        return await info.context.data_loaders.record_counts.load(
            (
                "trace",
                self.id,
                time_range or None,
                filter_condition or None,
                session_filter_condition or None,
            ),
        )

    @strawberry.field
    async def token_count_total(
        self,
        info: Info[Context, None],
        time_range: Optional[TimeRange] = UNSET,
        filter_condition: Optional[str] = UNSET,
    ) -> float:
        return await info.context.data_loaders.token_counts.load(
            ("total", self.id, time_range, filter_condition),
        )

    @strawberry.field
    async def token_count_prompt(
        self,
        info: Info[Context, None],
        time_range: Optional[TimeRange] = UNSET,
        filter_condition: Optional[str] = UNSET,
    ) -> float:
        return await info.context.data_loaders.token_counts.load(
            ("prompt", self.id, time_range, filter_condition),
        )

    @strawberry.field
    async def token_count_completion(
        self,
        info: Info[Context, None],
        time_range: Optional[TimeRange] = UNSET,
        filter_condition: Optional[str] = UNSET,
    ) -> float:
        return await info.context.data_loaders.token_counts.load(
            ("completion", self.id, time_range, filter_condition),
        )

    @strawberry.field
    async def cost_summary(
        self,
        info: Info[Context, None],
        time_range: Optional[TimeRange] = UNSET,
        filter_condition: Optional[str] = UNSET,
        session_filter_condition: Optional[str] = UNSET,
    ) -> SpanCostSummary:
        if filter_condition and session_filter_condition:
            raise BadRequest(
                "Both a filter condition and session filter condition "
                "cannot be applied at the same time"
            )
        summary = await info.context.data_loaders.span_cost_summary_by_project.load(
            (
                self.id,
                time_range or None,
                filter_condition or None,
                session_filter_condition or None,
            )
        )
        return SpanCostSummary(
            prompt=CostBreakdown(
                tokens=summary.prompt.tokens,
                cost=summary.prompt.cost,
            ),
            completion=CostBreakdown(
                tokens=summary.completion.tokens,
                cost=summary.completion.cost,
            ),
            total=CostBreakdown(
                tokens=summary.total.tokens,
                cost=summary.total.cost,
            ),
        )

    @strawberry.field
    async def latency_ms_quantile(
        self,
        info: Info[Context, None],
        probability: float,
        time_range: Optional[TimeRange] = UNSET,
        filter_condition: Optional[str] = UNSET,
        session_filter_condition: Optional[str] = UNSET,
    ) -> Optional[float]:
        if filter_condition and session_filter_condition:
            raise BadRequest(
                "Both a filter condition and session filter condition "
                "cannot be applied at the same time"
            )
        return await info.context.data_loaders.latency_ms_quantile.load(
            (
                "trace",
                self.id,
                time_range or None,
                filter_condition or None,
                session_filter_condition or None,
                probability,
            ),
        )

    @strawberry.field
    async def span_latency_ms_quantile(
        self,
        info: Info[Context, None],
        probability: float,
        time_range: Optional[TimeRange] = UNSET,
        filter_condition: Optional[str] = UNSET,
        session_filter_condition: Optional[str] = UNSET,
    ) -> Optional[float]:
        if filter_condition and session_filter_condition:
            raise BadRequest(
                "Both a filter condition and session filter condition "
                "cannot be applied at the same time"
            )
        return await info.context.data_loaders.latency_ms_quantile.load(
            (
                "span",
                self.id,
                time_range or None,
                filter_condition or None,
                session_filter_condition or None,
                probability,
            ),
        )

    @strawberry.field
    async def trace(self, trace_id: ID, info: Info[Context, None]) -> Optional[Trace]:
        stmt = (
            select(models.Trace)
            .where(models.Trace.trace_id == str(trace_id))
            .where(models.Trace.project_rowid == self.id)
        )
        async with info.context.db() as session:
            if (trace := await session.scalar(stmt)) is None:
                return None
        return Trace(id=trace.id, db_record=trace)

    @strawberry.field
    async def spans(
        self,
        info: Info[Context, None],
        time_range: Optional[TimeRange] = UNSET,
        first: Optional[int] = DEFAULT_PAGE_SIZE,
        last: Optional[int] = UNSET,
        after: Optional[CursorString] = UNSET,
        before: Optional[CursorString] = UNSET,
        sort: Optional[SpanSort] = UNSET,
        root_spans_only: Optional[bool] = UNSET,
        filter_condition: Optional[str] = UNSET,
        orphan_span_as_root_span: Optional[bool] = True,
    ) -> Connection[Span]:
        if root_spans_only and not filter_condition and sort and sort.col is SpanColumn.startTime:
            return await _paginate_span_by_trace_start_time(
                db=info.context.db,
                project_rowid=self.id,
                time_range=time_range,
                first=first,
                after=after,
                sort=sort,
                orphan_span_as_root_span=orphan_span_as_root_span,
            )
        stmt = (
            select(models.Span.id)
            .select_from(models.Span)
            .join(models.Trace)
            .where(models.Trace.project_rowid == self.id)
        )
        if time_range:
            if time_range.start:
                stmt = stmt.where(time_range.start <= models.Span.start_time)
            if time_range.end:
                stmt = stmt.where(models.Span.start_time < time_range.end)
        if filter_condition:
            span_filter = SpanFilter(condition=filter_condition)
            stmt = span_filter(stmt)
        sort_config: Optional[SpanSortConfig] = None
        cursor_rowid_column: Any = models.Span.id
        if sort:
            sort_config = sort.update_orm_expr(stmt)
            stmt = sort_config.stmt
            if sort_config.dir is SortDir.desc:
                cursor_rowid_column = desc(cursor_rowid_column)
        if after:
            cursor = Cursor.from_string(after)
            if sort_config and cursor.sort_column:
                sort_column = cursor.sort_column
                compare = operator.lt if sort_config.dir is SortDir.desc else operator.gt
                if sort_column.type is CursorSortColumnDataType.NULL:
                    stmt = stmt.where(sort_config.orm_expression.is_(None))
                    stmt = stmt.where(compare(models.Span.id, cursor.rowid))
                else:
                    stmt = stmt.where(
                        compare(
                            tuple_(sort_config.orm_expression, models.Span.id),
                            (sort_column.value, cursor.rowid),
                        )
                    )
            else:
                stmt = stmt.where(models.Span.id > cursor.rowid)
        stmt = stmt.order_by(cursor_rowid_column)
        if root_spans_only:
            # A root span is either a span with no parent_id or an orphan span
            # (a span whose parent_id references a span that doesn't exist in the database)
            if orphan_span_as_root_span:
                # Include both types of root spans
                parent_spans = select(models.Span.span_id).alias("parent_spans")
                candidate_spans = stmt.add_columns(models.Span.parent_id).cte("candidate_spans")
                stmt = select(candidate_spans).where(
                    or_(
                        candidate_spans.c.parent_id.is_(None),
                        ~select(1)
                        .where(candidate_spans.c.parent_id == parent_spans.c.span_id)
                        .exists(),
                    )
                )
            else:
                # Only include explicit root spans (spans with parent_id = NULL)
                stmt = stmt.where(models.Span.parent_id.is_(None))
        if first:
            stmt = stmt.limit(
                first + 1  # overfetch by one to determine whether there's a next page
            )
        cursors_and_nodes = []
        async with info.context.db() as session:
            span_records = await session.stream(stmt)
            async for span_record in islice(span_records, first):
                span_rowid: int = span_record[0]
                cursor = Cursor(rowid=span_rowid)
                if sort_config:
                    assert len(span_record) > 1
                    cursor.sort_column = CursorSortColumn(
                        type=sort_config.column_data_type,
                        value=span_record[1],
                    )
                cursors_and_nodes.append((cursor, Span(id=span_rowid)))
            has_next_page = True
            try:
                await span_records.__anext__()
            except StopAsyncIteration:
                has_next_page = False

        return connection_from_cursors_and_nodes(
            cursors_and_nodes,
            has_previous_page=False,
            has_next_page=has_next_page,
        )

    @strawberry.field
    async def sessions(
        self,
        info: Info[Context, None],
        time_range: Optional[TimeRange] = UNSET,
        first: Optional[int] = DEFAULT_PAGE_SIZE,
        after: Optional[CursorString] = UNSET,
        sort: Optional[ProjectSessionSort] = UNSET,
        filter_io_substring: Optional[str] = UNSET,
        session_id: Optional[str] = UNSET,
    ) -> Connection[ProjectSession]:
        table = models.ProjectSession
        if session_id:
            async with info.context.db() as session:
                ans = await session.scalar(
                    select(table).filter_by(
                        session_id=session_id,
                        project_id=self.id,
                    )
                )
            if ans:
                return connection_from_list(
                    data=[ProjectSession(id=ans.id, db_record=ans)],
                    args=ConnectionArgs(),
                )
            elif not filter_io_substring:
                return connection_from_list(
                    data=[],
                    args=ConnectionArgs(),
                )
        stmt = select(table).filter_by(project_id=self.id)
        if time_range:
            if time_range.start:
                stmt = stmt.where(time_range.start <= table.start_time)
            if time_range.end:
                stmt = stmt.where(table.start_time < time_range.end)
        if filter_io_substring:
            filtered_session_rowids = get_filtered_session_rowids_subquery(
                session_filter_condition=filter_io_substring,
                project_rowids=[self.id],
                start_time=time_range.start if time_range else None,
                end_time=time_range.end if time_range else None,
            )
            stmt = stmt.where(table.id.in_(filtered_session_rowids))
        sort_config: Optional[ProjectSessionSortConfig] = None
        cursor_rowid_column: Any = table.id
        if sort:
            sort_config = sort.update_orm_expr(stmt)
            stmt = sort_config.stmt
            if sort_config.dir is SortDir.desc:
                cursor_rowid_column = desc(cursor_rowid_column)
        if after:
            cursor = Cursor.from_string(after)
            if sort_config and cursor.sort_column:
                sort_column = cursor.sort_column
                compare = operator.lt if sort_config.dir is SortDir.desc else operator.gt
                if sort_column.type is CursorSortColumnDataType.NULL:
                    stmt = stmt.where(sort_config.orm_expression.is_(None))
                    stmt = stmt.where(compare(table.id, cursor.rowid))
                else:
                    stmt = stmt.where(
                        compare(
                            tuple_(sort_config.orm_expression, table.id),
                            (sort_column.value, cursor.rowid),
                        )
                    )
            else:
                stmt = stmt.where(table.id < cursor.rowid)
        stmt = stmt.order_by(cursor_rowid_column)
        if first:
            stmt = stmt.limit(
                first + 1  # over-fetch by one to determine whether there's a next page
            )
        cursors_and_nodes = []
        async with info.context.db() as session:
            records = await session.stream(stmt)
            async for record in islice(records, first):
                project_session = record[0]
                cursor = Cursor(rowid=project_session.id)
                if sort_config:
                    assert len(record) > 1
                    cursor.sort_column = CursorSortColumn(
                        type=sort_config.column_data_type,
                        value=record[1],
                    )
                cursors_and_nodes.append(
                    (cursor, ProjectSession(id=project_session.id, db_record=project_session))
                )
            has_next_page = True
            try:
                await records.__anext__()
            except StopAsyncIteration:
                has_next_page = False
        return connection_from_cursors_and_nodes(
            cursors_and_nodes,
            has_previous_page=False,
            has_next_page=has_next_page,
        )

    @strawberry.field(
        description="Names of all available annotations for traces. "
        "(The list contains no duplicates.)"
    )  # type: ignore
    async def trace_annotations_names(
        self,
        info: Info[Context, None],
    ) -> list[str]:
        stmt = (
            select(distinct(models.TraceAnnotation.name))
            .join(models.Trace)
            .where(models.Trace.project_rowid == self.id)
        )
        async with info.context.db() as session:
            return list(await session.scalars(stmt))

    @strawberry.field(
        description="Names of all available annotations for spans. "
        "(The list contains no duplicates.)"
    )  # type: ignore
    async def span_annotation_names(
        self,
        info: Info[Context, None],
    ) -> list[str]:
        stmt = (
            select(distinct(models.SpanAnnotation.name))
            .join(models.Span)
            .join(models.Trace, models.Span.trace_rowid == models.Trace.id)
            .where(models.Trace.project_rowid == self.id)
        )
        async with info.context.db() as session:
            return list(await session.scalars(stmt))

    @strawberry.field(
        description="Names of all available annotations for sessions. "
        "(The list contains no duplicates.)"
    )  # type: ignore
    async def session_annotation_names(
        self,
        info: Info[Context, None],
    ) -> list[str]:
        stmt = (
            select(distinct(models.ProjectSessionAnnotation.name))
            .join(models.ProjectSession)
            .where(models.ProjectSession.project_id == self.id)
        )
        async with info.context.db() as session:
            return list(await session.scalars(stmt))

    @strawberry.field(
        description="Names of available document evaluations.",
    )  # type: ignore
    async def document_evaluation_names(
        self,
        info: Info[Context, None],
        span_id: Optional[ID] = UNSET,
    ) -> list[str]:
        stmt = (
            select(distinct(models.DocumentAnnotation.name))
            .join(models.Span)
            .join(models.Trace, models.Span.trace_rowid == models.Trace.id)
            .where(models.Trace.project_rowid == self.id)
            .where(models.DocumentAnnotation.annotator_kind == "LLM")
        )
        if span_id:
            stmt = stmt.where(models.Span.span_id == str(span_id))
        async with info.context.db() as session:
            return list(await session.scalars(stmt))

    @strawberry.field
    async def trace_annotation_summary(
        self,
        info: Info[Context, None],
        annotation_name: str,
        filter_condition: Optional[str] = UNSET,
        session_filter_condition: Optional[str] = UNSET,
        time_range: Optional[TimeRange] = UNSET,
    ) -> Optional[AnnotationSummary]:
        if filter_condition and session_filter_condition:
            raise BadRequest(
                "Both a filter condition and session filter condition "
                "cannot be applied at the same time"
            )
        return await info.context.data_loaders.annotation_summaries.load(
            (
                "trace",
                self.id,
                time_range or None,
                filter_condition or None,
                session_filter_condition or None,
                annotation_name,
            ),
        )

    @strawberry.field
    async def span_annotation_summary(
        self,
        info: Info[Context, None],
        annotation_name: str,
        time_range: Optional[TimeRange] = UNSET,
        filter_condition: Optional[str] = UNSET,
        session_filter_condition: Optional[str] = UNSET,
    ) -> Optional[AnnotationSummary]:
        if filter_condition and session_filter_condition:
            raise BadRequest(
                "Both a filter condition and session filter condition "
                "cannot be applied at the same time"
            )
        return await info.context.data_loaders.annotation_summaries.load(
            (
                "span",
                self.id,
                time_range or None,
                filter_condition or None,
                session_filter_condition or None,
                annotation_name,
            ),
        )

    @strawberry.field
    async def document_evaluation_summary(
        self,
        info: Info[Context, None],
        evaluation_name: str,
        time_range: Optional[TimeRange] = UNSET,
        filter_condition: Optional[str] = UNSET,
    ) -> Optional[DocumentEvaluationSummary]:
        return await info.context.data_loaders.document_evaluation_summaries.load(
            (self.id, time_range, filter_condition, evaluation_name),
        )

    @strawberry.field
    def streaming_last_updated_at(
        self,
        info: Info[Context, None],
    ) -> Optional[datetime]:
        return info.context.last_updated_at.get(models.Project, self.id)

    @strawberry.field
    async def validate_span_filter_condition(
        self,
        info: Info[Context, None],
        condition: str,
    ) -> ValidationResult:
        """Validates a span filter condition by attempting to compile it for both SQLite and PostgreSQL.

        This method checks if the provided filter condition is syntactically valid and can be compiled
        into SQL queries for both SQLite and PostgreSQL databases. It does not execute the query,
        only validates its syntax. Any exception during compilation (syntax errors, invalid expressions,
        etc.) will result in an invalid validation result.

        Args:
            condition (str): The span filter condition string to validate.

        Returns:
            ValidationResult: A result object containing:
                - is_valid (bool): True if the condition is valid, False otherwise
                - error_message (Optional[str]): Error message if validation fails, None if valid
        """  # noqa: E501
        # This query is too expensive to run on every validation
        # valid_eval_names = await self.span_annotation_names()
        try:
            span_filter = SpanFilter(
                condition=condition,
                # valid_eval_names=valid_eval_names,
            )
            stmt = span_filter(select(models.Span))
            dialect = info.context.db.dialect
            if dialect is SupportedSQLDialect.POSTGRESQL:
                str(stmt.compile(dialect=sqlite.dialect()))
            elif dialect is SupportedSQLDialect.SQLITE:
                str(stmt.compile(dialect=postgresql.dialect()))  # type: ignore[no-untyped-call]
            else:
                assert_never(dialect)
            return ValidationResult(is_valid=True, error_message=None)
        except Exception as e:
            return ValidationResult(
                is_valid=False,
                error_message=str(e),
            )

    @strawberry.field
    async def annotation_configs(
        self,
        info: Info[Context, None],
        first: Optional[int] = 50,
        last: Optional[int] = None,
        after: Optional[str] = None,
        before: Optional[str] = None,
    ) -> Connection[AnnotationConfig]:
        args = ConnectionArgs(
            first=first,
            after=after if isinstance(after, CursorString) else None,
            last=last,
            before=before if isinstance(before, CursorString) else None,
        )
        loader = info.context.data_loaders.annotation_configs_by_project
        configs = await loader.load(self.id)
        data = [to_gql_annotation_config(config) for config in configs]
        return connection_from_list(data=data, args=args)

    @strawberry.field
    async def trace_retention_policy(
        self,
        info: Info[Context, None],
    ) -> Annotated["ProjectTraceRetentionPolicy", lazy(".ProjectTraceRetentionPolicy")]:
        from .ProjectTraceRetentionPolicy import ProjectTraceRetentionPolicy

        id_ = await info.context.data_loaders.trace_retention_policy_id_by_project_id.load(self.id)
        return ProjectTraceRetentionPolicy(id=id_)

    @strawberry.field
    async def created_at(
        self,
        info: Info[Context, None],
    ) -> datetime:
        if self.db_record:
            created_at = self.db_record.created_at
        else:
            created_at = await info.context.data_loaders.project_fields.load(
                (self.id, models.Project.created_at),
            )
        return created_at

    @strawberry.field
    async def updated_at(
        self,
        info: Info[Context, None],
    ) -> datetime:
        if self.db_record:
            updated_at = self.db_record.updated_at
        else:
            updated_at = await info.context.data_loaders.project_fields.load(
                (self.id, models.Project.updated_at),
            )
        return updated_at

    @strawberry.field
    async def span_count_time_series(
        self,
        info: Info[Context, None],
        time_range: TimeRange,
        time_bin_config: Optional[TimeBinConfig] = UNSET,
        filter_condition: Optional[str] = UNSET,
    ) -> "SpanCountTimeSeries":
        if time_range.start is None:
            raise BadRequest("Start time is required")

        dialect = info.context.db.dialect
        utc_offset_minutes = 0
        field: Literal["minute", "hour", "day", "week", "month", "year"] = "hour"
        if time_bin_config:
            utc_offset_minutes = time_bin_config.utc_offset_minutes
            if time_bin_config.scale is TimeBinScale.MINUTE:
                field = "minute"
            elif time_bin_config.scale is TimeBinScale.HOUR:
                field = "hour"
            elif time_bin_config.scale is TimeBinScale.DAY:
                field = "day"
            elif time_bin_config.scale is TimeBinScale.WEEK:
                field = "week"
            elif time_bin_config.scale is TimeBinScale.MONTH:
                field = "month"
            elif time_bin_config.scale is TimeBinScale.YEAR:
                field = "year"
        bucket = date_trunc(dialect, field, models.Span.start_time, utc_offset_minutes)
        stmt = (
            select(
                bucket,
                func.count(models.Span.id).label("total_count"),
                func.sum(case((models.Span.status_code == "OK", 1), else_=0)).label("ok_count"),
                func.sum(case((models.Span.status_code == "ERROR", 1), else_=0)).label(
                    "error_count"
                ),
                func.sum(case((models.Span.status_code == "UNSET", 1), else_=0)).label(
                    "unset_count"
                ),
            )
            .join_from(models.Span, models.Trace)
            .where(models.Trace.project_rowid == self.id)
            .group_by(bucket)
            .order_by(bucket)
        )
        if time_range.start:
            stmt = stmt.where(time_range.start <= models.Span.start_time)
        if time_range.end:
            stmt = stmt.where(models.Span.start_time < time_range.end)
        if filter_condition:
            span_filter = SpanFilter(condition=filter_condition)
            stmt = span_filter(stmt)

        data = {}
        async with info.context.db() as session:
            async for t, total_count, ok_count, error_count, unset_count in await session.stream(
                stmt
            ):
                timestamp = _as_datetime(t)
                data[timestamp] = SpanCountTimeSeriesDataPoint(
                    timestamp=timestamp,
                    ok_count=ok_count,
                    error_count=error_count,
                    unset_count=unset_count,
                    total_count=total_count,
                )

        data_timestamps: list[datetime] = [data_point.timestamp for data_point in data.values()]
        min_time = min([*data_timestamps, time_range.start])
        max_time = max(
            [
                *data_timestamps,
                *([time_range.end] if time_range.end else [datetime.now(timezone.utc)]),
            ],
        )
        for timestamp in get_timestamp_range(
            start_time=min_time,
            end_time=max_time,
            stride=field,
            utc_offset_minutes=utc_offset_minutes,
        ):
            if timestamp not in data:
                data[timestamp] = SpanCountTimeSeriesDataPoint(timestamp=timestamp)
        return SpanCountTimeSeries(data=sorted(data.values(), key=lambda x: x.timestamp))

    @strawberry.field
    async def trace_count_time_series(
        self,
        info: Info[Context, None],
        time_range: TimeRange,
        time_bin_config: Optional[TimeBinConfig] = UNSET,
    ) -> "TraceCountTimeSeries":
        if time_range.start is None:
            raise BadRequest("Start time is required")

        dialect = info.context.db.dialect
        utc_offset_minutes = 0
        field: Literal["minute", "hour", "day", "week", "month", "year"] = "hour"
        if time_bin_config:
            utc_offset_minutes = time_bin_config.utc_offset_minutes
            if time_bin_config.scale is TimeBinScale.MINUTE:
                field = "minute"
            elif time_bin_config.scale is TimeBinScale.HOUR:
                field = "hour"
            elif time_bin_config.scale is TimeBinScale.DAY:
                field = "day"
            elif time_bin_config.scale is TimeBinScale.WEEK:
                field = "week"
            elif time_bin_config.scale is TimeBinScale.MONTH:
                field = "month"
            elif time_bin_config.scale is TimeBinScale.YEAR:
                field = "year"
        bucket = date_trunc(dialect, field, models.Trace.start_time, utc_offset_minutes)
        stmt = (
            select(bucket, func.count(models.Trace.id))
            .where(models.Trace.project_rowid == self.id)
            .group_by(bucket)
            .order_by(bucket)
        )
        if time_range:
            if time_range.start:
                stmt = stmt.where(time_range.start <= models.Trace.start_time)
            if time_range.end:
                stmt = stmt.where(models.Trace.start_time < time_range.end)
        data = {}
        async with info.context.db() as session:
            async for t, v in await session.stream(stmt):
                timestamp = _as_datetime(t)
                data[timestamp] = TimeSeriesDataPoint(timestamp=timestamp, value=v)

        data_timestamps: list[datetime] = [data_point.timestamp for data_point in data.values()]
        min_time = min([*data_timestamps, time_range.start])
        max_time = max(
            [
                *data_timestamps,
                *([time_range.end] if time_range.end else [datetime.now(timezone.utc)]),
            ],
        )
        for timestamp in get_timestamp_range(
            start_time=min_time,
            end_time=max_time,
            stride=field,
            utc_offset_minutes=utc_offset_minutes,
        ):
            if timestamp not in data:
                data[timestamp] = TimeSeriesDataPoint(timestamp=timestamp)
        return TraceCountTimeSeries(data=sorted(data.values(), key=lambda x: x.timestamp))

    @strawberry.field
    async def trace_count_by_status_time_series(
        self,
        info: Info[Context, None],
        time_range: TimeRange,
        time_bin_config: Optional[TimeBinConfig] = UNSET,
    ) -> "TraceCountByStatusTimeSeries":
        if time_range.start is None:
            raise BadRequest("Start time is required")

        dialect = info.context.db.dialect
        utc_offset_minutes = 0
        field: Literal["minute", "hour", "day", "week", "month", "year"] = "hour"
        if time_bin_config:
            utc_offset_minutes = time_bin_config.utc_offset_minutes
            if time_bin_config.scale is TimeBinScale.MINUTE:
                field = "minute"
            elif time_bin_config.scale is TimeBinScale.HOUR:
                field = "hour"
            elif time_bin_config.scale is TimeBinScale.DAY:
                field = "day"
            elif time_bin_config.scale is TimeBinScale.WEEK:
                field = "week"
            elif time_bin_config.scale is TimeBinScale.MONTH:
                field = "month"
            elif time_bin_config.scale is TimeBinScale.YEAR:
                field = "year"
        bucket = date_trunc(dialect, field, models.Trace.start_time, utc_offset_minutes)
        trace_error_status_counts = (
            select(
                models.Span.trace_rowid,
            )
            .where(models.Span.parent_id.is_(None))
            .group_by(models.Span.trace_rowid)
            .having(func.max(models.Span.cumulative_error_count) > 0)
        ).subquery()
        stmt = (
            select(
                bucket,
                func.count(models.Trace.id).label("total_count"),
                func.coalesce(func.count(trace_error_status_counts.c.trace_rowid), 0).label(
                    "error_count"
                ),
            )
            .join_from(
                models.Trace,
                trace_error_status_counts,
                onclause=trace_error_status_counts.c.trace_rowid == models.Trace.id,
                isouter=True,
            )
            .where(models.Trace.project_rowid == self.id)
            .group_by(bucket)
            .order_by(bucket)
        )
        if time_range:
            if time_range.start:
                stmt = stmt.where(time_range.start <= models.Trace.start_time)
            if time_range.end:
                stmt = stmt.where(models.Trace.start_time < time_range.end)
        data: dict[datetime, TraceCountByStatusTimeSeriesDataPoint] = {}
        async with info.context.db() as session:
            async for t, total_count, error_count in await session.stream(stmt):
                timestamp = _as_datetime(t)
                data[timestamp] = TraceCountByStatusTimeSeriesDataPoint(
                    timestamp=timestamp,
                    ok_count=total_count - error_count,
                    error_count=error_count,
                    total_count=total_count,
                )

        data_timestamps: list[datetime] = [data_point.timestamp for data_point in data.values()]
        min_time = min([*data_timestamps, time_range.start])
        max_time = max(
            [
                *data_timestamps,
                *([time_range.end] if time_range.end else [datetime.now(timezone.utc)]),
            ],
        )
        for timestamp in get_timestamp_range(
            start_time=min_time,
            end_time=max_time,
            stride=field,
            utc_offset_minutes=utc_offset_minutes,
        ):
            if timestamp not in data:
                data[timestamp] = TraceCountByStatusTimeSeriesDataPoint(
                    timestamp=timestamp,
                    ok_count=0,
                    error_count=0,
                    total_count=0,
                )
        return TraceCountByStatusTimeSeries(data=sorted(data.values(), key=lambda x: x.timestamp))

    @strawberry.field
    async def trace_latency_ms_percentile_time_series(
        self,
        info: Info[Context, None],
        time_range: TimeRange,
        time_bin_config: Optional[TimeBinConfig] = UNSET,
    ) -> "TraceLatencyPercentileTimeSeries":
        if time_range.start is None:
            raise BadRequest("Start time is required")

        dialect = info.context.db.dialect
        utc_offset_minutes = 0
        field: Literal["minute", "hour", "day", "week", "month", "year"] = "hour"
        if time_bin_config:
            utc_offset_minutes = time_bin_config.utc_offset_minutes
            if time_bin_config.scale is TimeBinScale.MINUTE:
                field = "minute"
            elif time_bin_config.scale is TimeBinScale.HOUR:
                field = "hour"
            elif time_bin_config.scale is TimeBinScale.DAY:
                field = "day"
            elif time_bin_config.scale is TimeBinScale.WEEK:
                field = "week"
            elif time_bin_config.scale is TimeBinScale.MONTH:
                field = "month"
            elif time_bin_config.scale is TimeBinScale.YEAR:
                field = "year"
        bucket = date_trunc(dialect, field, models.Trace.start_time, utc_offset_minutes)

        stmt = select(bucket).where(models.Trace.project_rowid == self.id)
        if time_range.start:
            stmt = stmt.where(time_range.start <= models.Trace.start_time)
        if time_range.end:
            stmt = stmt.where(models.Trace.start_time < time_range.end)

        if dialect is SupportedSQLDialect.POSTGRESQL:
            stmt = stmt.add_columns(
                percentile_cont(0.50).within_group(models.Trace.latency_ms.asc()).label("p50"),
                percentile_cont(0.75).within_group(models.Trace.latency_ms.asc()).label("p75"),
                percentile_cont(0.90).within_group(models.Trace.latency_ms.asc()).label("p90"),
                percentile_cont(0.95).within_group(models.Trace.latency_ms.asc()).label("p95"),
                percentile_cont(0.99).within_group(models.Trace.latency_ms.asc()).label("p99"),
                percentile_cont(0.999).within_group(models.Trace.latency_ms.asc()).label("p999"),
                func.max(models.Trace.latency_ms).label("max"),
            )
        elif dialect is SupportedSQLDialect.SQLITE:
            stmt = stmt.add_columns(
                func.percentile(models.Trace.latency_ms, 50).label("p50"),
                func.percentile(models.Trace.latency_ms, 75).label("p75"),
                func.percentile(models.Trace.latency_ms, 90).label("p90"),
                func.percentile(models.Trace.latency_ms, 95).label("p95"),
                func.percentile(models.Trace.latency_ms, 99).label("p99"),
                func.percentile(models.Trace.latency_ms, 99.9).label("p999"),
                func.max(models.Trace.latency_ms).label("max"),
            )
        else:
            assert_never(dialect)

        stmt = stmt.group_by(bucket).order_by(bucket)

        data: dict[datetime, TraceLatencyMsPercentileTimeSeriesDataPoint] = {}
        async with info.context.db() as session:
            async for (
                bucket_time,
                p50,
                p75,
                p90,
                p95,
                p99,
                p999,
                max_latency,
            ) in await session.stream(stmt):
                timestamp = _as_datetime(bucket_time)
                data[timestamp] = TraceLatencyMsPercentileTimeSeriesDataPoint(
                    timestamp=timestamp,
                    p50=p50,
                    p75=p75,
                    p90=p90,
                    p95=p95,
                    p99=p99,
                    p999=p999,
                    max=max_latency,
                )

        data_timestamps: list[datetime] = [data_point.timestamp for data_point in data.values()]
        min_time = min([*data_timestamps, time_range.start])
        max_time = max(
            [
                *data_timestamps,
                *([time_range.end] if time_range.end else [datetime.now(timezone.utc)]),
            ],
        )
        for timestamp in get_timestamp_range(
            start_time=min_time,
            end_time=max_time,
            stride=field,
            utc_offset_minutes=utc_offset_minutes,
        ):
            if timestamp not in data:
                data[timestamp] = TraceLatencyMsPercentileTimeSeriesDataPoint(timestamp=timestamp)
        return TraceLatencyPercentileTimeSeries(
            data=sorted(data.values(), key=lambda x: x.timestamp)
        )

    @strawberry.field
    async def trace_token_count_time_series(
        self,
        info: Info[Context, None],
        time_range: TimeRange,
        time_bin_config: Optional[TimeBinConfig] = UNSET,
    ) -> "TraceTokenCountTimeSeries":
        if time_range.start is None:
            raise BadRequest("Start time is required")

        dialect = info.context.db.dialect
        utc_offset_minutes = 0
        field: Literal["minute", "hour", "day", "week", "month", "year"] = "hour"
        if time_bin_config:
            utc_offset_minutes = time_bin_config.utc_offset_minutes
            if time_bin_config.scale is TimeBinScale.MINUTE:
                field = "minute"
            elif time_bin_config.scale is TimeBinScale.HOUR:
                field = "hour"
            elif time_bin_config.scale is TimeBinScale.DAY:
                field = "day"
            elif time_bin_config.scale is TimeBinScale.WEEK:
                field = "week"
            elif time_bin_config.scale is TimeBinScale.MONTH:
                field = "month"
            elif time_bin_config.scale is TimeBinScale.YEAR:
                field = "year"
        bucket = date_trunc(dialect, field, models.Trace.start_time, utc_offset_minutes)
        stmt = (
            select(
                bucket,
                func.sum(models.SpanCost.total_tokens),
                func.sum(models.SpanCost.prompt_tokens),
                func.sum(models.SpanCost.completion_tokens),
            )
            .join_from(
                models.Trace,
                models.SpanCost,
                onclause=models.SpanCost.trace_rowid == models.Trace.id,
            )
            .where(models.Trace.project_rowid == self.id)
            .group_by(bucket)
            .order_by(bucket)
        )
        if time_range:
            if time_range.start:
                stmt = stmt.where(time_range.start <= models.Trace.start_time)
            if time_range.end:
                stmt = stmt.where(models.Trace.start_time < time_range.end)
        data: dict[datetime, TraceTokenCountTimeSeriesDataPoint] = {}
        async with info.context.db() as session:
            async for (
                t,
                total_tokens,
                prompt_tokens,
                completion_tokens,
            ) in await session.stream(stmt):
                timestamp = _as_datetime(t)
                data[timestamp] = TraceTokenCountTimeSeriesDataPoint(
                    timestamp=timestamp,
                    prompt_token_count=prompt_tokens,
                    completion_token_count=completion_tokens,
                    total_token_count=total_tokens,
                )

        data_timestamps: list[datetime] = [data_point.timestamp for data_point in data.values()]
        min_time = min([*data_timestamps, time_range.start])
        max_time = max(
            [
                *data_timestamps,
                *([time_range.end] if time_range.end else [datetime.now(timezone.utc)]),
            ],
        )
        for timestamp in get_timestamp_range(
            start_time=min_time,
            end_time=max_time,
            stride=field,
            utc_offset_minutes=utc_offset_minutes,
        ):
            if timestamp not in data:
                data[timestamp] = TraceTokenCountTimeSeriesDataPoint(timestamp=timestamp)
        return TraceTokenCountTimeSeries(data=sorted(data.values(), key=lambda x: x.timestamp))

    @strawberry.field
    async def trace_token_cost_time_series(
        self,
        info: Info[Context, None],
        time_range: TimeRange,
        time_bin_config: Optional[TimeBinConfig] = UNSET,
    ) -> "TraceTokenCostTimeSeries":
        if time_range.start is None:
            raise BadRequest("Start time is required")

        dialect = info.context.db.dialect
        utc_offset_minutes = 0
        field: Literal["minute", "hour", "day", "week", "month", "year"] = "hour"
        if time_bin_config:
            utc_offset_minutes = time_bin_config.utc_offset_minutes
            if time_bin_config.scale is TimeBinScale.MINUTE:
                field = "minute"
            elif time_bin_config.scale is TimeBinScale.HOUR:
                field = "hour"
            elif time_bin_config.scale is TimeBinScale.DAY:
                field = "day"
            elif time_bin_config.scale is TimeBinScale.WEEK:
                field = "week"
            elif time_bin_config.scale is TimeBinScale.MONTH:
                field = "month"
            elif time_bin_config.scale is TimeBinScale.YEAR:
                field = "year"
        bucket = date_trunc(dialect, field, models.Trace.start_time, utc_offset_minutes)
        stmt = (
            select(
                bucket,
                func.sum(models.SpanCost.total_cost),
                func.sum(models.SpanCost.prompt_cost),
                func.sum(models.SpanCost.completion_cost),
            )
            .join_from(
                models.Trace,
                models.SpanCost,
                onclause=models.SpanCost.trace_rowid == models.Trace.id,
            )
            .where(models.Trace.project_rowid == self.id)
            .group_by(bucket)
            .order_by(bucket)
        )
        if time_range:
            if time_range.start:
                stmt = stmt.where(time_range.start <= models.Trace.start_time)
            if time_range.end:
                stmt = stmt.where(models.Trace.start_time < time_range.end)
        data: dict[datetime, TraceTokenCostTimeSeriesDataPoint] = {}
        async with info.context.db() as session:
            async for (
                t,
                total_cost,
                prompt_cost,
                completion_cost,
            ) in await session.stream(stmt):
                timestamp = _as_datetime(t)
                data[timestamp] = TraceTokenCostTimeSeriesDataPoint(
                    timestamp=timestamp,
                    prompt_cost=prompt_cost,
                    completion_cost=completion_cost,
                    total_cost=total_cost,
                )

        data_timestamps: list[datetime] = [data_point.timestamp for data_point in data.values()]
        min_time = min([*data_timestamps, time_range.start])
        max_time = max(
            [
                *data_timestamps,
                *([time_range.end] if time_range.end else [datetime.now(timezone.utc)]),
            ],
        )
        for timestamp in get_timestamp_range(
            start_time=min_time,
            end_time=max_time,
            stride=field,
            utc_offset_minutes=utc_offset_minutes,
        ):
            if timestamp not in data:
                data[timestamp] = TraceTokenCostTimeSeriesDataPoint(timestamp=timestamp)
        return TraceTokenCostTimeSeries(data=sorted(data.values(), key=lambda x: x.timestamp))

    @strawberry.field
    async def span_annotation_score_time_series(
        self,
        info: Info[Context, None],
        time_range: TimeRange,
        time_bin_config: Optional[TimeBinConfig] = UNSET,
    ) -> "SpanAnnotationScoreTimeSeries":
        if time_range.start is None:
            raise BadRequest("Start time is required")

        dialect = info.context.db.dialect
        utc_offset_minutes = 0
        field: Literal["minute", "hour", "day", "week", "month", "year"] = "hour"
        if time_bin_config:
            utc_offset_minutes = time_bin_config.utc_offset_minutes
            if time_bin_config.scale is TimeBinScale.MINUTE:
                field = "minute"
            elif time_bin_config.scale is TimeBinScale.HOUR:
                field = "hour"
            elif time_bin_config.scale is TimeBinScale.DAY:
                field = "day"
            elif time_bin_config.scale is TimeBinScale.WEEK:
                field = "week"
            elif time_bin_config.scale is TimeBinScale.MONTH:
                field = "month"
            elif time_bin_config.scale is TimeBinScale.YEAR:
                field = "year"
        bucket = date_trunc(dialect, field, models.Trace.start_time, utc_offset_minutes)
        stmt = (
            select(
                bucket,
                models.SpanAnnotation.name,
                func.avg(models.SpanAnnotation.score).label("average_score"),
            )
            .join_from(
                models.SpanAnnotation,
                models.Span,
                onclause=models.SpanAnnotation.span_rowid == models.Span.id,
            )
            .join_from(
                models.Span,
                models.Trace,
                onclause=models.Span.trace_rowid == models.Trace.id,
            )
            .where(models.Trace.project_rowid == self.id)
            .group_by(bucket, models.SpanAnnotation.name)
            .order_by(bucket)
        )
        if time_range:
            if time_range.start:
                stmt = stmt.where(time_range.start <= models.Trace.start_time)
            if time_range.end:
                stmt = stmt.where(models.Trace.start_time < time_range.end)
        scores: dict[datetime, dict[str, float]] = {}
        unique_names: set[str] = set()
        async with info.context.db() as session:
            async for (
                t,
                name,
                average_score,
            ) in await session.stream(stmt):
                if average_score is None:
                    continue
                timestamp = _as_datetime(t)
                if timestamp not in scores:
                    scores[timestamp] = {}
                scores[timestamp][name] = average_score
                unique_names.add(name)

        score_timestamps: list[datetime] = [timestamp for timestamp in scores]
        min_time = min([*score_timestamps, time_range.start])
        max_time = max(
            [
                *score_timestamps,
                *([time_range.end] if time_range.end else [datetime.now(timezone.utc)]),
            ],
        )
        data: dict[datetime, SpanAnnotationScoreTimeSeriesDataPoint] = {
            timestamp: SpanAnnotationScoreTimeSeriesDataPoint(
                timestamp=timestamp,
                scores_with_labels=[
                    SpanAnnotationScoreWithLabel(label=label, score=scores[timestamp][label])
                    for label in scores[timestamp]
                ],
            )
            for timestamp in score_timestamps
        }
        for timestamp in get_timestamp_range(
            start_time=min_time,
            end_time=max_time,
            stride=field,
            utc_offset_minutes=utc_offset_minutes,
        ):
            if timestamp not in data:
                data[timestamp] = SpanAnnotationScoreTimeSeriesDataPoint(
                    timestamp=timestamp,
                    scores_with_labels=[],
                )
        return SpanAnnotationScoreTimeSeries(
            data=sorted(data.values(), key=lambda x: x.timestamp),
            names=sorted(list(unique_names)),
        )

    @strawberry.field
    async def top_models_by_cost(
        self,
        info: Info[Context, None],
        time_range: TimeRange,
    ) -> list[GenerativeModel]:
        if time_range.start is None:
            raise BadRequest("Start time is required")

        async with info.context.db() as session:
            stmt = (
                select(
                    models.GenerativeModel,
                    func.sum(models.SpanCost.total_tokens).label("total_tokens"),
                    func.sum(models.SpanCost.prompt_tokens).label("prompt_tokens"),
                    func.sum(models.SpanCost.completion_tokens).label("completion_tokens"),
                    func.sum(models.SpanCost.total_cost).label("total_cost"),
                    func.sum(models.SpanCost.prompt_cost).label("prompt_cost"),
                    func.sum(models.SpanCost.completion_cost).label("completion_cost"),
                )
                .join(
                    models.SpanCost,
                    models.SpanCost.model_id == models.GenerativeModel.id,
                )
                .join(
                    models.Trace,
                    models.SpanCost.trace_rowid == models.Trace.id,
                )
                .where(models.Trace.project_rowid == self.id)
                .where(models.SpanCost.model_id.isnot(None))
                .where(models.SpanCost.span_start_time >= time_range.start)
                .group_by(models.GenerativeModel.id)
                .order_by(func.sum(models.SpanCost.total_cost).desc())
            )
            if time_range.end:
                stmt = stmt.where(models.SpanCost.span_start_time < time_range.end)
            results: list[GenerativeModel] = []
            async for (
                model,
                total_tokens,
                prompt_tokens,
                completion_tokens,
                total_cost,
                prompt_cost,
                completion_cost,
            ) in await session.stream(stmt):
                cost_summary = SpanCostSummary(
                    prompt=CostBreakdown(tokens=prompt_tokens, cost=prompt_cost),
                    completion=CostBreakdown(tokens=completion_tokens, cost=completion_cost),
                    total=CostBreakdown(tokens=total_tokens, cost=total_cost),
                )
                cache_time_range = TimeRange(
                    start=time_range.start,
                    end=time_range.end,
                )
                gql_model = GenerativeModel(id=model.id, db_record=model)
                gql_model.add_cached_cost_summary(self.id, cache_time_range, cost_summary)
                results.append(gql_model)
            return results

    @strawberry.field
    async def top_models_by_token_count(
        self,
        info: Info[Context, None],
        time_range: TimeRange,
    ) -> list[GenerativeModel]:
        if time_range.start is None:
            raise BadRequest("Start time is required")

        async with info.context.db() as session:
            stmt = (
                select(
                    models.GenerativeModel,
                    func.sum(models.SpanCost.total_tokens).label("total_tokens"),
                    func.sum(models.SpanCost.prompt_tokens).label("prompt_tokens"),
                    func.sum(models.SpanCost.completion_tokens).label("completion_tokens"),
                    func.sum(models.SpanCost.total_cost).label("total_cost"),
                    func.sum(models.SpanCost.prompt_cost).label("prompt_cost"),
                    func.sum(models.SpanCost.completion_cost).label("completion_cost"),
                )
                .join(
                    models.SpanCost,
                    models.SpanCost.model_id == models.GenerativeModel.id,
                )
                .join(
                    models.Trace,
                    models.SpanCost.trace_rowid == models.Trace.id,
                )
                .where(models.Trace.project_rowid == self.id)
                .where(models.SpanCost.model_id.isnot(None))
                .where(models.SpanCost.span_start_time >= time_range.start)
                .group_by(models.GenerativeModel.id)
                .order_by(func.sum(models.SpanCost.total_tokens).desc())
            )
            if time_range.end:
                stmt = stmt.where(models.SpanCost.span_start_time < time_range.end)
            results: list[GenerativeModel] = []
            async for (
                model,
                total_tokens,
                prompt_tokens,
                completion_tokens,
                total_cost,
                prompt_cost,
                completion_cost,
            ) in await session.stream(stmt):
                cost_summary = SpanCostSummary(
                    prompt=CostBreakdown(tokens=prompt_tokens, cost=prompt_cost),
                    completion=CostBreakdown(tokens=completion_tokens, cost=completion_cost),
                    total=CostBreakdown(tokens=total_tokens, cost=total_cost),
                )
                cache_time_range = TimeRange(
                    start=time_range.start,
                    end=time_range.end,
                )
                gql_model = GenerativeModel(id=model.id, db_record=model)
                gql_model.add_cached_cost_summary(self.id, cache_time_range, cost_summary)
                results.append(gql_model)
            return results


@strawberry.type
class SpanCountTimeSeriesDataPoint:
    timestamp: datetime
    ok_count: Optional[int] = None
    error_count: Optional[int] = None
    unset_count: Optional[int] = None
    total_count: Optional[int] = None


@strawberry.type
class SpanCountTimeSeries:
    data: list[SpanCountTimeSeriesDataPoint]


@strawberry.type
class TraceCountTimeSeries(TimeSeries):
    """A time series of trace count"""


@strawberry.type
class TraceCountByStatusTimeSeriesDataPoint:
    timestamp: datetime
    ok_count: int
    error_count: int
    total_count: int


@strawberry.type
class TraceCountByStatusTimeSeries:
    data: list[TraceCountByStatusTimeSeriesDataPoint]


@strawberry.type
class TraceLatencyMsPercentileTimeSeriesDataPoint:
    timestamp: datetime
    p50: Optional[float] = None
    p75: Optional[float] = None
    p90: Optional[float] = None
    p95: Optional[float] = None
    p99: Optional[float] = None
    p999: Optional[float] = None
    max: Optional[float] = None


@strawberry.type
class TraceLatencyPercentileTimeSeries:
    data: list[TraceLatencyMsPercentileTimeSeriesDataPoint]


@strawberry.type
class TraceTokenCountTimeSeriesDataPoint:
    timestamp: datetime
    prompt_token_count: Optional[float] = None
    completion_token_count: Optional[float] = None
    total_token_count: Optional[float] = None


@strawberry.type
class TraceTokenCountTimeSeries:
    data: list[TraceTokenCountTimeSeriesDataPoint]


@strawberry.type
class TraceTokenCostTimeSeriesDataPoint:
    timestamp: datetime
    prompt_cost: Optional[float] = None
    completion_cost: Optional[float] = None
    total_cost: Optional[float] = None


@strawberry.type
class TraceTokenCostTimeSeries:
    data: list[TraceTokenCostTimeSeriesDataPoint]


@strawberry.type
class SpanAnnotationScoreWithLabel:
    label: str
    score: float


@strawberry.type
class SpanAnnotationScoreTimeSeriesDataPoint:
    timestamp: datetime
    scores_with_labels: list[SpanAnnotationScoreWithLabel]


@strawberry.type
class SpanAnnotationScoreTimeSeries:
    data: list[SpanAnnotationScoreTimeSeriesDataPoint]
    names: list[str]


INPUT_VALUE = SpanAttributes.INPUT_VALUE.split(".")
OUTPUT_VALUE = SpanAttributes.OUTPUT_VALUE.split(".")


def _as_datetime(value: Any) -> datetime:
    if isinstance(value, datetime):
        return value
    if isinstance(value, str):
        return cast(datetime, normalize_datetime(datetime.fromisoformat(value), timezone.utc))
    raise ValueError(f"Cannot convert {value} to datetime")


async def _paginate_span_by_trace_start_time(
    db: DbSessionFactory,
    project_rowid: int,
    time_range: Optional[TimeRange] = None,
    first: Optional[int] = DEFAULT_PAGE_SIZE,
    after: Optional[CursorString] = None,
    sort: SpanSort = SpanSort(col=SpanColumn.startTime, dir=SortDir.desc),
    orphan_span_as_root_span: Optional[bool] = True,
    retries: int = 3,
) -> Connection[Span]:
    """Return one representative root span per trace, ordered by trace start time.

    **Note**: Despite the function name, cursors are based on trace rowids, not span rowids.
    This is because we paginate by traces (one span per trace), not individual spans.

    **Important**: The edges list can be empty while has_next_page=True. This happens
    when traces exist but have no matching root spans. Pagination continues because there
    may be more traces ahead with spans.

    Args:
        db: Database session factory.
        project_rowid: Project ID to query spans from.
        time_range: Optional time range filter on trace start times.
        first: Maximum number of edges to return (default: DEFAULT_PAGE_SIZE).
        after: Cursor for pagination (points to trace position, not span).
        sort: Sort by trace start time (asc/desc only).
        orphan_span_as_root_span: Whether to include orphan spans as root spans.
            True: spans with parent_id=NULL OR pointing to non-existent spans.
            False: only spans with parent_id=NULL.
        retries: Maximum number of retry attempts when insufficient edges are found.
            When traces exist but lack root spans, the function retries pagination
            to find traces with spans. Set to 0 to disable retries.

    Returns:
        Connection[Span] with:
        - edges: At most one Edge per trace (may be empty list).
        - page_info: Pagination info based on trace positions.

    Key Points:
        - Traces without root spans produce NO edges
        - Spans ordered by trace start time, not span start time
        - Cursors track trace positions for efficient large-scale pagination
    """
    # Build base trace query ordered by start time
    traces = select(
        models.Trace.id,
        models.Trace.start_time,
    ).where(models.Trace.project_rowid == project_rowid)
    if sort.dir is SortDir.desc:
        traces = traces.order_by(
            models.Trace.start_time.desc(),
            models.Trace.id.desc(),
        )
    else:
        traces = traces.order_by(
            models.Trace.start_time.asc(),
            models.Trace.id.asc(),
        )

    # Apply time range filters
    if time_range:
        if time_range.start:
            traces = traces.where(time_range.start <= models.Trace.start_time)
        if time_range.end:
            traces = traces.where(models.Trace.start_time < time_range.end)

    # Apply cursor pagination
    if after:
        cursor = Cursor.from_string(after)
        assert cursor.sort_column
        compare = operator.lt if sort.dir is SortDir.desc else operator.gt
        traces = traces.where(
            compare(
                tuple_(models.Trace.start_time, models.Trace.id),
                (cursor.sort_column.value, cursor.rowid),
            )
        )

    # Limit for pagination
    if first:
        traces = traces.limit(
            first + 1  # over-fetch by one to determine whether there's a next page
        )
    traces_cte = traces.cte()

    # Define join condition for root spans
    if orphan_span_as_root_span:
        # Include both NULL parent_id and orphaned spans
        parent_spans = select(models.Span.span_id).alias("parent_spans")
        onclause = and_(
            models.Span.trace_rowid == traces_cte.c.id,
            or_(
                models.Span.parent_id.is_(None),
                ~exists().where(models.Span.parent_id == parent_spans.c.span_id),
            ),
        )
    else:
        # Only spans with no parent (parent_id is NULL, excludes orphaned spans)
        onclause = and_(
            models.Span.trace_rowid == traces_cte.c.id,
            models.Span.parent_id.is_(None),
        )

    # Join traces with root spans (left join allows traces without spans)
    stmt = select(
        traces_cte.c.id,
        traces_cte.c.start_time,
        models.Span.id,
    ).join_from(
        traces_cte,
        models.Span,
        onclause=onclause,
        isouter=True,
    )

    # Order by trace time, then pick earliest span per trace
    if sort.dir is SortDir.desc:
        stmt = stmt.order_by(
            traces_cte.c.start_time.desc(),
            traces_cte.c.id.desc(),
            models.Span.start_time.asc(),  # earliest span
            models.Span.id.desc(),
        )
    else:
        stmt = stmt.order_by(
            traces_cte.c.start_time.asc(),
            traces_cte.c.id.asc(),
            models.Span.start_time.asc(),  # earliest span
            models.Span.id.desc(),
        )

    # Use DISTINCT for PostgreSQL, manual grouping for SQLite
    if db.dialect is SupportedSQLDialect.POSTGRESQL:
        stmt = stmt.distinct(traces_cte.c.start_time, traces_cte.c.id)
    elif db.dialect is SupportedSQLDialect.SQLITE:
        # too complicated for SQLite, so we rely on groupby() below
        pass
    else:
        assert_never(db.dialect)

    # Process results and build edges
    edges: list[Edge[Span]] = []
    start_cursor: Optional[str] = None
    end_cursor: Optional[str] = None
    async with db() as session:
        records = groupby(await session.stream(stmt), key=lambda record: record[:2])
        async for (trace_rowid, trace_start_time), group in islice(records, first):
            cursor = Cursor(
                rowid=trace_rowid,
                sort_column=CursorSortColumn(
                    type=CursorSortColumnDataType.DATETIME,
                    value=trace_start_time,
                ),
            )
            if start_cursor is None:
                start_cursor = str(cursor)
            end_cursor = str(cursor)
            first_record = group[0]
            # Only create edge if trace has a root span
            if (span_rowid := first_record[2]) is not None:
                edges.append(Edge(node=Span(id=span_rowid), cursor=str(cursor)))
        has_next_page = True
        try:
            await records.__anext__()
        except StopAsyncIteration:
            has_next_page = False

    # Retry if we need more edges and more traces exist
    if first and len(edges) < first and has_next_page:
        while retries and (num_needed := first - len(edges)) and has_next_page:
            retries -= 1
            batch_size = max(first, 1000)
            more = await _paginate_span_by_trace_start_time(
                db=db,
                project_rowid=project_rowid,
                time_range=time_range,
                first=batch_size,
                after=end_cursor,
                sort=sort,
                orphan_span_as_root_span=orphan_span_as_root_span,
                retries=0,
            )
            edges.extend(more.edges[:num_needed])
            start_cursor = start_cursor or more.page_info.start_cursor
            end_cursor = more.page_info.end_cursor if len(edges) < first else edges[-1].cursor
            has_next_page = len(more.edges) > num_needed or more.page_info.has_next_page

    return Connection(
        edges=edges,
        page_info=PageInfo(
            start_cursor=start_cursor,
            end_cursor=end_cursor,
            has_previous_page=False,
            has_next_page=has_next_page,
        ),
    )


def to_gql_project(project: models.Project) -> Project:
    """
    Converts an ORM project to a GraphQL project.
    """
    return Project(
        id=project.id,
        db_record=project,
    )



from collections import defaultdict
from typing import Optional

from sqlalchemy import select
from strawberry.dataloader import DataLoader
from typing_extensions import TypeAlias

from phoenix.db import models
from phoenix.server.types import DbSessionFactory

UserRoleId: TypeAlias = int
Key: TypeAlias = UserRoleId
Result: TypeAlias = Optional[models.UserRole]


class UserRolesDataLoader(DataLoader[Key, Result]):
    """DataLoader that batches together user roles by their ids."""

    def __init__(self, db: DbSessionFactory) -> None:
        super().__init__(load_fn=self._load_fn)
        self._db = db

    async def _load_fn(self, keys: list[Key]) -> list[Result]:
        user_roles_by_id: defaultdict[Key, Result] = defaultdict(None)
        async with self._db() as session:
            data = await session.stream_scalars(select(models.UserRole))
            async for user_role in data:
                user_roles_by_id[user_role.id] = user_role

        return [user_roles_by_id.get(role_id) for role_id in keys]



from __future__ import annotations

import logging

logger = logging.getLogger(__name__)


def generate_aws_rds_token(
    host: str,
    port: int,
    user: str,
) -> str:
    """Generate an AWS RDS IAM authentication token.

    This function creates a short-lived (15 minutes) authentication token for connecting
    to AWS RDS/Aurora PostgreSQL instances using IAM database authentication.

    The AWS region is automatically resolved using boto3.

    Args:
        host: The database hostname (e.g., 'mydb.abc123.us-west-2.rds.amazonaws.com')
        port: The database port (typically 5432 for PostgreSQL)
        user: The database username (must match an IAM-enabled database user)

    Returns:
        A temporary authentication token string to use as the database password

    Raises:
        ImportError: If boto3 is not installed
        Exception: If AWS credentials/region are not configured or token generation fails

    Example:
        >>> token = generate_aws_rds_token(
        ...     host='mydb.us-west-2.rds.amazonaws.com',
        ...     port=5432,
        ...     user='myuser'
        ... )
    """
    try:
        import boto3  # type: ignore
    except ImportError as e:
        raise ImportError(
            "boto3 is required for AWS RDS IAM authentication. "
            "Install it with: pip install 'arize-phoenix[aws]'"
        ) from e

    try:
        client = boto3.client("rds")

        logger.debug(f"Generating AWS RDS IAM auth token for user '{user}' at {host}:{port}")
        token = client.generate_db_auth_token(  # pyright: ignore
            DBHostname=host,
            Port=port,
            DBUsername=user,
        )

        return str(token)  # pyright: ignore

    except Exception as e:
        logger.error(
            f"Failed to generate AWS RDS IAM authentication token: {e}. "
            "Ensure AWS credentials are configured and have 'rds-db:connect' permission."
        )
        raise



import re
from datetime import datetime, timezone
from typing import Optional

import sqlalchemy as sa
import strawberry
from sqlalchemy import delete
from sqlalchemy.exc import IntegrityError as PostgreSQLIntegrityError
from sqlalchemy.orm import joinedload
from sqlean.dbapi2 import IntegrityError as SQLiteIntegrityError  # type: ignore[import-untyped]
from strawberry.relay import GlobalID
from strawberry.types import Info

from phoenix.db import models
from phoenix.server.api.auth import IsNotReadOnly, IsNotViewer
from phoenix.server.api.context import Context
from phoenix.server.api.exceptions import BadRequest, Conflict, NotFound
from phoenix.server.api.queries import Query
from phoenix.server.api.types.GenerativeModel import GenerativeModel
from phoenix.server.api.types.node import from_global_id_with_expected_type
from phoenix.server.api.types.TokenPrice import TokenKind


@strawberry.input
class TokenPriceInput:
    token_type: str
    cost_per_million_tokens: float
    kind: TokenKind

    @property
    def token_prices(self) -> models.TokenPrice:
        """Generate TokenPrice instances based on the input."""
        return models.TokenPrice(
            token_type=self.token_type,
            is_prompt=self.kind == TokenKind.PROMPT,
            base_rate=self.cost_per_million_tokens / 1_000_000,
        )


@strawberry.input
class CreateModelMutationInput:
    name: str
    provider: Optional[str] = None
    name_pattern: str
    costs: list[TokenPriceInput]
    start_time: Optional[datetime] = None


@strawberry.type
class CreateModelMutationPayload:
    model: GenerativeModel
    query: Query


@strawberry.input
class UpdateModelMutationInput:
    id: GlobalID
    name: str
    provider: Optional[str]
    name_pattern: str
    costs: list[TokenPriceInput]
    start_time: Optional[datetime] = None


@strawberry.type
class UpdateModelMutationPayload:
    model: GenerativeModel
    query: Query


@strawberry.input
class DeleteModelMutationInput:
    id: GlobalID


@strawberry.type
class DeleteModelMutationPayload:
    model: GenerativeModel
    query: Query


@strawberry.type
class ModelMutationMixin:
    @strawberry.mutation(permission_classes=[IsNotReadOnly, IsNotViewer])  # type: ignore
    async def create_model(
        self,
        info: Info[Context, None],
        input: CreateModelMutationInput,
    ) -> CreateModelMutationPayload:
        cost_types = set(cost.token_type for cost in input.costs)
        if "input" not in cost_types:
            raise BadRequest("input cost is required")
        if "output" not in cost_types:
            raise BadRequest("output cost is required")
        name_pattern = _compile_regular_expression(input.name_pattern)
        token_prices = [cost.token_prices for cost in input.costs]
        model = models.GenerativeModel(
            name=input.name,
            provider=input.provider,
            name_pattern=name_pattern,
            is_built_in=False,
            token_prices=token_prices,
            start_time=input.start_time,
        )
        async with info.context.db() as session:
            session.add(model)
            try:
                await session.flush()
            except (PostgreSQLIntegrityError, SQLiteIntegrityError):
                raise Conflict(f"Model with name '{input.name}' already exists")

        return CreateModelMutationPayload(
            model=GenerativeModel(id=model.id, db_record=model),
            query=Query(),
        )

    @strawberry.mutation(permission_classes=[IsNotReadOnly, IsNotViewer])  # type: ignore
    async def update_model(
        self,
        info: Info[Context, None],
        input: UpdateModelMutationInput,
    ) -> UpdateModelMutationPayload:
        try:
            model_id = from_global_id_with_expected_type(input.id, GenerativeModel.__name__)
        except ValueError:
            raise BadRequest(f'Invalid model id: "{input.id}"')

        cost_types = set(cost.token_type for cost in input.costs)
        if "input" not in cost_types:
            raise BadRequest("input cost is required")
        if "output" not in cost_types:
            raise BadRequest("output cost is required")
        name_pattern = _compile_regular_expression(input.name_pattern)
        token_prices = [cost.token_prices for cost in input.costs]
        async with info.context.db() as session:
            model = await session.scalar(
                sa.select(models.GenerativeModel)
                .where(models.GenerativeModel.deleted_at.is_(None))
                .where(models.GenerativeModel.id == model_id)
                .options(joinedload(models.GenerativeModel.token_prices))
            )
            if model is None:
                raise NotFound(f'Model "{input.id}" not found')
            if model.is_built_in:
                raise BadRequest("Cannot update built-in model")

            await session.execute(
                delete(models.TokenPrice).where(models.TokenPrice.model_id == model.id)
            )

            await session.refresh(model)

            model.name = input.name
            model.provider = input.provider or ""
            model.name_pattern = name_pattern
            model.token_prices = token_prices
            model.start_time = input.start_time
            # Explicitly set updated_at so the GenerativeModelStore daemon picks up this
            # change (SQLAlchemy's onupdate may not trigger for relationship-only changes).
            model.updated_at = datetime.now(timezone.utc)
            session.add(model)
            try:
                await session.flush()
            except (PostgreSQLIntegrityError, SQLiteIntegrityError):
                raise Conflict(f"Model with name '{input.name}' already exists")

        return UpdateModelMutationPayload(
            model=GenerativeModel(id=model.id, db_record=model),
            query=Query(),
        )

    @strawberry.mutation(permission_classes=[IsNotReadOnly, IsNotViewer])  # type: ignore
    async def delete_model(
        self,
        info: Info[Context, None],
        input: DeleteModelMutationInput,
    ) -> DeleteModelMutationPayload:
        try:
            model_id = from_global_id_with_expected_type(input.id, GenerativeModel.__name__)
        except ValueError:
            raise BadRequest(f'Invalid model id: "{input.id}"')

        async with info.context.db() as session:
            model = await session.scalar(
                sa.update(models.GenerativeModel)
                .values(deleted_at=datetime.now(timezone.utc))
                .where(models.GenerativeModel.deleted_at.is_(None))
                .where(models.GenerativeModel.id == model_id)
                .returning(models.GenerativeModel)
            )
            if model is None:
                raise NotFound(f'Model "{input.id}" not found')
            if model.is_built_in:
                await session.rollback()
                raise BadRequest("Cannot delete built-in model")
        return DeleteModelMutationPayload(
            model=GenerativeModel(id=model.id, db_record=model),
            query=Query(),
        )


def _compile_regular_expression(maybe_regex: str) -> re.Pattern[str]:
    """
    Compile the given string as a regular expression.
    Raises a BadRequest error if the given string is not a valid regex.
    """
    try:
        return re.compile(maybe_regex)
    except re.error as error:
        raise BadRequest(f"Invalid regex: {str(error)}")



from typing import Any, Generic, Optional, TypedDict, TypeVar, Union

from fastapi import HTTPException
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession
from strawberry.relay import GlobalID
from typing_extensions import TypeAlias, assert_never

from phoenix.db import models
from phoenix.server.api.types.node import from_global_id_with_expected_type
from phoenix.server.api.types.Project import Project as ProjectNodeType

from .models import V1RoutesBaseModel

StatusCode: TypeAlias = int
DataType = TypeVar("DataType")
Responses: TypeAlias = dict[
    Union[int, str], dict[str, Any]
]  # input type for the `responses` parameter of a fastapi route


class StatusCodeWithDescription(TypedDict):
    """
    A duck type for a status code with a description detailing under what
    conditions the status code is raised.
    """

    status_code: StatusCode
    description: str


class RequestBody(V1RoutesBaseModel, Generic[DataType]):
    # A generic request type accepted by V1 routes.
    #
    # Don't use """ for this docstring or it will be included as a description
    # in the generated OpenAPI schema.
    data: DataType


class ResponseBody(V1RoutesBaseModel, Generic[DataType]):
    # A generic response type returned by V1 routes.
    #
    # Don't use """ for this docstring or it will be included as a description
    # in the generated OpenAPI schema.

    data: DataType


class PaginatedResponseBody(V1RoutesBaseModel, Generic[DataType]):
    # A generic paginated response type returned by V1 routes.
    #
    # Don't use """ for this docstring or it will be included as a description
    # in the generated OpenAPI schema.

    data: list[DataType]
    next_cursor: Optional[str]


def add_errors_to_responses(
    errors: list[Union[StatusCode, StatusCodeWithDescription]],
    /,
    *,
    responses: Optional[Responses] = None,
) -> Responses:
    """
    Creates or updates a patch for an OpenAPI schema's `responses` section to
    include status codes in the generated OpenAPI schema.
    """
    output_responses: Responses = responses or {}
    for error in errors:
        status_code: int
        description: Optional[str] = None
        if isinstance(error, StatusCode):
            status_code = error
        elif isinstance(error, dict):
            status_code = error["status_code"]
            description = error["description"]
        else:
            assert_never(error)
        if status_code not in output_responses:
            output_responses[status_code] = {
                "content": {"text/plain": {"schema": {"type": "string"}}}
            }
        if description:
            output_responses[status_code]["description"] = description
    return output_responses


def add_text_csv_content_to_responses(
    status_code: StatusCode, /, *, responses: Optional[Responses] = None
) -> Responses:
    """
    Creates or updates a patch for an OpenAPI schema's `responses` section to
    ensure that the response for the given status code is marked as text/csv in
    the generated OpenAPI schema.
    """
    output_responses: Responses = responses or {}
    if status_code not in output_responses:
        output_responses[status_code] = {}
    output_responses[status_code]["content"] = {
        "text/csv": {"schema": {"type": "string", "contentMediaType": "text/csv"}}
    }
    return output_responses


async def _get_project_by_identifier(
    session: AsyncSession,
    project_identifier: str,
) -> models.Project:
    """
    Get a project by its ID or name.

    Args:
        session: The database session.
        project_identifier: The project ID or name.

    Returns:
        The project object.

    Raises:
        HTTPException: If the identifier format is invalid or the project is not found.
    """
    # Try to parse as a GlobalID first
    try:
        id_ = from_global_id_with_expected_type(
            GlobalID.from_id(project_identifier),
            ProjectNodeType.__name__,
        )
    except Exception:
        try:
            name = project_identifier
        except HTTPException:
            raise HTTPException(
                status_code=422,
                detail=f"Invalid project identifier format: {project_identifier}",
            )
        stmt = select(models.Project).filter_by(name=name)
        project = await session.scalar(stmt)
        if project is None:
            raise HTTPException(
                status_code=404,
                detail=f"Project with name {name} not found",
            )
    else:
        project = await session.get(models.Project, id_)
        if project is None:
            raise HTTPException(
                status_code=404,
                detail=f"Project with ID {project_identifier} not found",
            )
    return project



from collections import defaultdict

from sqlalchemy import select
from strawberry.dataloader import DataLoader
from typing_extensions import TypeAlias

from phoenix.db import models
from phoenix.server.types import DbSessionFactory

Key: TypeAlias = int
Result: TypeAlias = list[models.DocumentAnnotation]


class DocumentEvaluationsDataLoader(DataLoader[Key, Result]):
    def __init__(self, db: DbSessionFactory) -> None:
        super().__init__(load_fn=self._load_fn)
        self._db = db

    async def _load_fn(self, keys: list[Key]) -> list[Result]:
        document_annotations_by_id: defaultdict[Key, Result] = defaultdict(list)
        mda = models.DocumentAnnotation
        async with self._db() as session:
            data = await session.stream_scalars(select(mda).where(mda.span_rowid.in_(keys)))
            async for document_evaluation in data:
                document_annotations_by_id[document_evaluation.span_rowid].append(
                    document_evaluation
                )
        return [document_annotations_by_id[key] for key in keys]



from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, Optional

import strawberry
from openinference.semconv.trace import OpenInferenceLLMProviderValues
from strawberry.relay import Node, NodeID
from strawberry.relay.types import GlobalID
from strawberry.types import Info
from strawberry.types.unset import UNSET
from typing_extensions import TypeAlias, assert_never

from phoenix.db import models
from phoenix.server.api.context import Context
from phoenix.server.api.exceptions import BadRequest
from phoenix.server.api.input_types.TimeRange import TimeRange
from phoenix.server.api.types.CostBreakdown import CostBreakdown
from phoenix.server.api.types.GenerativeProvider import GenerativeProviderKey
from phoenix.server.api.types.ModelInterface import ModelInterface
from phoenix.server.api.types.node import from_global_id
from phoenix.server.api.types.SpanCostDetailSummaryEntry import SpanCostDetailSummaryEntry
from phoenix.server.api.types.SpanCostSummary import SpanCostSummary
from phoenix.server.api.types.TokenPrice import TokenKind, TokenPrice


@strawberry.enum
class GenerativeModelKind(Enum):
    CUSTOM = "CUSTOM"
    BUILT_IN = "BUILT_IN"


ProjectId: TypeAlias = int
TimeRangeKey: TypeAlias = tuple[Optional[datetime], Optional[datetime]]
CachedCostSummaryKey: TypeAlias = tuple[Optional[ProjectId], TimeRangeKey]


@strawberry.type
class GenerativeModel(Node, ModelInterface):
    id: NodeID[int]
    db_record: strawberry.Private[Optional[models.GenerativeModel]] = None
    cached_cost_summary: strawberry.Private[
        Optional[dict[CachedCostSummaryKey, SpanCostSummary]]
    ] = None

    def __post_init__(self) -> None:
        if self.db_record and self.id != self.db_record.id:
            raise ValueError("GenerativeModel ID mismatch")

    @strawberry.field
    async def name(self, info: Info[Context, None]) -> str:
        if self.db_record:
            val = self.db_record.name
        else:
            val = await info.context.data_loaders.generative_model_fields.load(
                (self.id, models.GenerativeModel.name),
            )
        return val

    @strawberry.field
    async def provider(self, info: Info[Context, None]) -> Optional[str]:
        if self.db_record:
            provider = self.db_record.provider
        else:
            provider = await info.context.data_loaders.generative_model_fields.load(
                (self.id, models.GenerativeModel.provider),
            )
        return provider or None

    @strawberry.field
    async def name_pattern(self, info: Info[Context, None]) -> str:
        if self.db_record:
            pattern = self.db_record.name_pattern.pattern
        else:
            name_pattern_obj = await info.context.data_loaders.generative_model_fields.load(
                (self.id, models.GenerativeModel.name_pattern),
            )
            pattern = name_pattern_obj.pattern
        assert isinstance(pattern, str)
        return pattern

    @strawberry.field
    async def kind(self, info: Info[Context, None]) -> GenerativeModelKind:
        if self.db_record:
            is_built_in = self.db_record.is_built_in
        else:
            is_built_in = await info.context.data_loaders.generative_model_fields.load(
                (self.id, models.GenerativeModel.is_built_in),
            )
        return GenerativeModelKind.BUILT_IN if is_built_in else GenerativeModelKind.CUSTOM

    @strawberry.field
    async def created_at(self, info: Info[Context, None]) -> datetime:
        if self.db_record:
            val = self.db_record.created_at
        else:
            val = await info.context.data_loaders.generative_model_fields.load(
                (self.id, models.GenerativeModel.created_at),
            )
        return val

    @strawberry.field
    async def updated_at(self, info: Info[Context, None]) -> datetime:
        if self.db_record:
            val = self.db_record.updated_at
        else:
            val = await info.context.data_loaders.generative_model_fields.load(
                (self.id, models.GenerativeModel.updated_at),
            )
        return val

    @strawberry.field
    async def provider_key(self, info: Info[Context, None]) -> Optional[GenerativeProviderKey]:
        if self.db_record:
            provider = self.db_record.provider
        else:
            provider = await info.context.data_loaders.generative_model_fields.load(
                (self.id, models.GenerativeModel.provider),
            )
        return _semconv_provider_to_gql_generative_provider_key(provider) if provider else None

    @strawberry.field
    async def start_time(self, info: Info[Context, None]) -> Optional[datetime]:
        if self.db_record:
            val = self.db_record.start_time
        else:
            val = await info.context.data_loaders.generative_model_fields.load(
                (self.id, models.GenerativeModel.start_time),
            )
        return val

    def add_cached_cost_summary(
        self, project_id: Optional[int], time_range: TimeRange, cost_summary: SpanCostSummary
    ) -> None:
        if self.cached_cost_summary is None:
            self.cached_cost_summary = {}
        time_range_key = (time_range.start, time_range.end) if time_range else (None, None)
        cache_key = (project_id, time_range_key)
        self.cached_cost_summary[cache_key] = cost_summary

    @strawberry.field
    async def token_prices(self, info: Info[Context, None]) -> list[TokenPrice]:
        costs = await info.context.data_loaders.token_prices_by_model.load(self.id)
        token_prices: list[TokenPrice] = []
        for cost in costs:
            token_prices.append(
                TokenPrice(
                    token_type=cost.token_type,
                    kind=TokenKind.PROMPT if cost.is_prompt else TokenKind.COMPLETION,
                    cost_per_million_tokens=cost.base_rate * 1_000_000,
                    cost_per_token=cost.base_rate,
                )
            )
        return token_prices

    @strawberry.field
    async def cost_summary(
        self,
        info: Info[Context, None],
        project_id: Optional[GlobalID] = UNSET,
        time_range: Optional[TimeRange] = UNSET,
    ) -> SpanCostSummary:
        if self.cached_cost_summary is not None:
            time_range_key = (time_range.start, time_range.end) if time_range else (None, None)
            project_rowid: Optional[int] = None
            if project_id:
                type_name, project_rowid = from_global_id(project_id)
                if type_name != models.Project.__name__:
                    raise BadRequest("Invalid Project ID")
            cache_key = (project_rowid, time_range_key)
            if cache_key in self.cached_cost_summary:
                return self.cached_cost_summary[cache_key]

        if time_range or project_id:
            raise BadRequest(
                "Cost summaries for specific projects or time ranges are not yet implemented"
            )

        loader = info.context.data_loaders.span_cost_summary_by_generative_model
        summary = await loader.load(self.id)
        return SpanCostSummary(
            prompt=CostBreakdown(
                tokens=summary.prompt.tokens,
                cost=summary.prompt.cost,
            ),
            completion=CostBreakdown(
                tokens=summary.completion.tokens,
                cost=summary.completion.cost,
            ),
            total=CostBreakdown(
                tokens=summary.total.tokens,
                cost=summary.total.cost,
            ),
        )

    @strawberry.field
    async def cost_detail_summary_entries(
        self,
        info: Info[Context, None],
    ) -> list[SpanCostDetailSummaryEntry]:
        loader = info.context.data_loaders.span_cost_detail_summary_entries_by_generative_model
        summary = await loader.load(self.id)
        return [
            SpanCostDetailSummaryEntry(
                token_type=entry.token_type,
                is_prompt=entry.is_prompt,
                value=CostBreakdown(
                    tokens=entry.value.tokens,
                    cost=entry.value.cost,
                ),
            )
            for entry in summary
        ]

    @strawberry.field
    async def last_used_at(self, info: Info[Context, None]) -> Optional[datetime]:
        return await info.context.data_loaders.last_used_times_by_generative_model_id.load(self.id)


def _semconv_provider_to_gql_generative_provider_key(
    semconv_provider_str: str,
) -> Optional[GenerativeProviderKey]:
    """
    Translates a semconv provider string to a GQL GenerativeProviderKey.
    """

    try:
        semconv_provider = OpenInferenceLLMProviderValues(semconv_provider_str)
    except Exception:
        return None
    if semconv_provider == OpenInferenceLLMProviderValues.OPENAI:
        return GenerativeProviderKey.OPENAI
    if semconv_provider == OpenInferenceLLMProviderValues.ANTHROPIC:
        return GenerativeProviderKey.ANTHROPIC
    if semconv_provider == OpenInferenceLLMProviderValues.AZURE:
        return GenerativeProviderKey.AZURE_OPENAI
    if semconv_provider == OpenInferenceLLMProviderValues.GOOGLE:
        return GenerativeProviderKey.GOOGLE
    if semconv_provider == OpenInferenceLLMProviderValues.DEEPSEEK:
        return GenerativeProviderKey.DEEPSEEK
    if semconv_provider == OpenInferenceLLMProviderValues.XAI:
        return GenerativeProviderKey.XAI
    if semconv_provider == OpenInferenceLLMProviderValues.AWS:
        return GenerativeProviderKey.AWS
    if semconv_provider == OpenInferenceLLMProviderValues.COHERE:
        return None  # TODO
    if semconv_provider == OpenInferenceLLMProviderValues.MISTRALAI:
        return None  # TODO
    if TYPE_CHECKING:
        assert_never(semconv_provider)
    else:
        return None



"""Add prompt tables

Revision ID: bc8fea3c2bc8
Revises: 4ded9e43755f
Create Date: 2024-12-16 15:45:01.090563

"""

from typing import Any, Sequence, Union

import sqlalchemy as sa
from alembic import op
from sqlalchemy import JSON
from sqlalchemy.dialects import postgresql
from sqlalchemy.ext.compiler import compiles


class JSONB(JSON):
    # See https://docs.sqlalchemy.org/en/20/core/custom_types.html
    __visit_name__ = "JSONB"


@compiles(JSONB, "sqlite")
def _(*args: Any, **kwargs: Any) -> str:
    # See https://docs.sqlalchemy.org/en/20/core/custom_types.html
    return "JSONB"


JSON_ = (
    JSON()
    .with_variant(
        postgresql.JSONB(),
        "postgresql",
    )
    .with_variant(
        JSONB(),
        "sqlite",
    )
)

revision: str = "bc8fea3c2bc8"
down_revision: Union[str, None] = "4ded9e43755f"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.create_table(
        "prompt_labels",
        sa.Column("id", sa.Integer, primary_key=True),
        sa.Column("name", sa.String, nullable=False, unique=True, index=True),
        sa.Column("description", sa.String, nullable=True),
        sa.Column("color", sa.String, nullable=True),
    )

    op.create_table(
        "prompts",
        sa.Column("id", sa.Integer, primary_key=True),
        sa.Column(
            "source_prompt_id",
            sa.Integer,
            sa.ForeignKey("prompts.id", ondelete="SET NULL"),
            nullable=True,
            index=True,
        ),
        sa.Column("name", sa.String, unique=True, index=True, nullable=False),
        sa.Column("description", sa.String, nullable=True),
        sa.Column("metadata", JSON_, nullable=False),
        sa.Column(
            "created_at",
            sa.TIMESTAMP(timezone=True),
            nullable=False,
            server_default=sa.func.now(),
        ),
        sa.Column(
            "updated_at",
            sa.TIMESTAMP(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
            onupdate=sa.func.now(),
        ),
    )

    op.create_table(
        "prompts_prompt_labels",
        sa.Column("id", sa.Integer, primary_key=True),
        sa.Column(
            "prompt_label_id",
            sa.Integer,
            sa.ForeignKey("prompt_labels.id", ondelete="CASCADE"),
            nullable=False,
            index=True,
        ),
        sa.Column(
            "prompt_id",
            sa.Integer,
            sa.ForeignKey("prompts.id", ondelete="CASCADE"),
            nullable=False,
            index=True,
        ),
        sa.UniqueConstraint(
            "prompt_label_id",
            "prompt_id",
        ),
    )

    op.create_table(
        "prompt_versions",
        sa.Column("id", sa.Integer, primary_key=True),
        sa.Column(
            "prompt_id",
            sa.Integer,
            sa.ForeignKey("prompts.id", ondelete="CASCADE"),
            nullable=False,
            index=True,
        ),
        sa.Column("description", sa.String, nullable=True),
        sa.Column(
            "user_id",
            sa.Integer,
            sa.ForeignKey("users.id", ondelete="SET NULL"),
            nullable=True,
            index=True,
        ),
        sa.Column(
            "template_type",
            sa.String,
            sa.CheckConstraint(
                "template_type IN ('CHAT', 'STR')",
                name="template_type",
            ),
            nullable=False,
        ),
        sa.Column(
            "template_format",
            sa.String,
            sa.CheckConstraint(
                "template_format IN ('F_STRING', 'MUSTACHE', 'NONE')",
                name="template_format",
            ),
            nullable=False,
        ),
        sa.Column("template", JSON_, nullable=False),
        sa.Column("invocation_parameters", JSON_, nullable=False),
        sa.Column("tools", JSON_, nullable=True),
        sa.Column("response_format", JSON_, nullable=True),
        sa.Column("model_provider", sa.String, nullable=False),
        sa.Column("model_name", sa.String, nullable=False),
        sa.Column("metadata", JSON_, nullable=False),
        sa.Column(
            "created_at",
            sa.TIMESTAMP(timezone=True),
            nullable=False,
            server_default=sa.func.now(),
        ),
        sqlite_autoincrement=True,
    )

    op.create_table(
        "prompt_version_tags",
        sa.Column("id", sa.Integer, primary_key=True),
        sa.Column("name", sa.String, nullable=False),
        sa.Column("description", sa.String, nullable=True),
        sa.Column(
            "prompt_id",
            sa.Integer,
            sa.ForeignKey("prompts.id", ondelete="CASCADE"),
            nullable=False,
            index=True,
        ),
        sa.Column(
            "prompt_version_id",
            sa.Integer,
            sa.ForeignKey("prompt_versions.id", ondelete="CASCADE"),
            nullable=False,
            index=True,
        ),
        sa.Column(
            "user_id",
            sa.Integer,
            sa.ForeignKey("users.id", ondelete="SET NULL"),
            nullable=True,
            index=True,
        ),
        sa.UniqueConstraint(
            "name",
            "prompt_id",
        ),
    )


def downgrade() -> None:
    op.drop_table("prompt_version_tags")
    op.drop_table("prompt_versions")
    op.drop_table("prompts_prompt_labels")
    op.drop_table("prompts")
    op.drop_table("prompt_labels")



from sqlalchemy import select
from strawberry.dataloader import DataLoader
from typing_extensions import TypeAlias

from phoenix.db import models
from phoenix.server.types import DbSessionFactory

DatasetID: TypeAlias = int
Key: TypeAlias = DatasetID
Result: TypeAlias = list[models.DatasetSplit]


class DatasetDatasetSplitsDataLoader(DataLoader[Key, Result]):
    def __init__(self, db: DbSessionFactory) -> None:
        super().__init__(
            load_fn=self._load_fn,
        )
        self._db = db

    async def _load_fn(self, keys: list[Key]) -> list[Result]:
        dataset_ids = keys
        async with self._db() as session:
            splits: dict[DatasetID, dict[int, models.DatasetSplit]] = {
                dataset_id: {} for dataset_id in dataset_ids
            }

            async for dataset_id, split in await session.stream(
                select(models.DatasetExample.dataset_id, models.DatasetSplit)
                .select_from(models.DatasetSplit)
                .join(
                    models.DatasetSplitDatasetExample,
                    onclause=(
                        models.DatasetSplit.id == models.DatasetSplitDatasetExample.dataset_split_id
                    ),
                )
                .join(
                    models.DatasetExample,
                    onclause=(
                        models.DatasetSplitDatasetExample.dataset_example_id
                        == models.DatasetExample.id
                    ),
                )
                .where(models.DatasetExample.dataset_id.in_(dataset_ids))
            ):
                # Use dict to deduplicate splits by split.id
                if dataset_id in splits:
                    splits[dataset_id][split.id] = split

            return [
                sorted(splits.get(dataset_id, {}).values(), key=lambda x: x.name)
                for dataset_id in keys
            ]



import asyncio
import logging
import secrets
from datetime import datetime, timedelta, timezone
from functools import partial
from typing import TYPE_CHECKING
from urllib.parse import urlencode, urlparse, urlunparse

from fastapi import APIRouter, Depends, HTTPException, Request, Response
from sqlalchemy import func, select
from sqlalchemy.orm import joinedload

from phoenix.auth import (
    DEFAULT_SECRET_LENGTH,
    PHOENIX_ACCESS_TOKEN_COOKIE_NAME,
    PHOENIX_REFRESH_TOKEN_COOKIE_NAME,
    Token,
    compute_password_hash,
    delete_access_token_cookie,
    delete_oauth2_nonce_cookie,
    delete_oauth2_state_cookie,
    delete_refresh_token_cookie,
    is_valid_password,
    sanitize_email,
    set_access_token_cookie,
    set_refresh_token_cookie,
    validate_password_format,
)
from phoenix.config import (
    get_base_url,
    get_env_disable_basic_auth,
    get_env_disable_rate_limit,
)
from phoenix.db import models
from phoenix.server.api.routers.ldap import get_or_create_ldap_user
from phoenix.server.bearer_auth import PhoenixUser, create_access_and_refresh_tokens
from phoenix.server.email.types import EmailSender
from phoenix.server.rate_limiters import ServerRateLimiter, fastapi_ip_rate_limiter
from phoenix.server.types import (
    AccessTokenClaims,
    PasswordResetTokenClaims,
    PasswordResetTokenId,
    RefreshTokenClaims,
    TokenStore,
    UserId,
)
from phoenix.server.utils import prepend_root_path

if TYPE_CHECKING:
    from phoenix.server.ldap import LDAPAuthenticator

logger = logging.getLogger(__name__)

rate_limiter = ServerRateLimiter(
    per_second_rate_limit=0.2,
    enforcement_window_seconds=60,
    partition_seconds=60,
    active_partitions=2,
)


def create_auth_router(ldap_enabled: bool = False) -> APIRouter:
    """Create auth router with all authentication endpoints.

    Creates a fresh router instance each time to avoid global state issues
    (e.g., route accumulation in tests).

    Security: Only registers the /ldap/login endpoint when LDAP is actually configured.
    This prevents information disclosure and reduces attack surface.

    Args:
        ldap_enabled: Whether LDAP authentication is configured

    Returns:
        APIRouter: Authentication router with all endpoints registered
    """
    # Build rate limiter paths based on configuration
    rate_limited_paths = [
        "/auth/login",
        "/auth/logout",
        "/auth/refresh",
        "/auth/password-reset-email",
        "/auth/password-reset",
    ]
    if ldap_enabled:
        rate_limited_paths.append("/auth/ldap/login")

    login_rate_limiter = fastapi_ip_rate_limiter(rate_limiter, paths=rate_limited_paths)
    auth_dependencies = [Depends(login_rate_limiter)] if not get_env_disable_rate_limit() else []

    router = APIRouter(prefix="/auth", include_in_schema=False, dependencies=auth_dependencies)

    # Register all authentication endpoints
    router.add_api_route("/login", _login, methods=["POST"])
    router.add_api_route("/logout", _logout, methods=["GET"])
    router.add_api_route("/refresh", _refresh_tokens, methods=["POST"])
    router.add_api_route("/password-reset-email", _initiate_password_reset, methods=["POST"])
    router.add_api_route("/password-reset", _reset_password, methods=["POST"])

    # Conditionally add LDAP endpoint only if configured
    if ldap_enabled:
        router.add_api_route("/ldap/login", _ldap_login, methods=["POST"])

    return router


async def _login(request: Request) -> Response:
    """Authenticate user via email/password and return access/refresh tokens."""
    if get_env_disable_basic_auth():
        raise HTTPException(status_code=403)
    data = await request.json()
    email = data.get("email")
    password = data.get("password")

    if not email or not password:
        raise HTTPException(status_code=401, detail="Email and password required")

    # Sanitize email by trimming and lowercasing
    email = sanitize_email(email)

    async with request.app.state.db() as session:
        user = await session.scalar(
            select(models.User)
            .where(func.lower(models.User.email) == email)
            .options(joinedload(models.User.role))
        )
        if (
            user is None
            or (password_hash := user.password_hash) is None
            or (salt := user.password_salt) is None
        ):
            raise HTTPException(status_code=401, detail=LOGIN_FAILED_MESSAGE)

    loop = asyncio.get_running_loop()
    password_is_valid = partial(
        is_valid_password, password=password, salt=salt, password_hash=password_hash
    )
    if not await loop.run_in_executor(None, password_is_valid):
        raise HTTPException(status_code=401, detail=LOGIN_FAILED_MESSAGE)

    return await _create_auth_response(request, user)


async def _logout(request: Request) -> Response:
    """Log out user by revoking tokens and clearing cookies."""
    token_store: TokenStore = request.app.state.get_token_store()
    user_id = None
    if isinstance(user := request.user, PhoenixUser):
        user_id = user.identity
    elif (refresh_token := request.cookies.get(PHOENIX_REFRESH_TOKEN_COOKIE_NAME)) and (
        isinstance(
            refresh_token_claims := await token_store.read(Token(refresh_token)),
            RefreshTokenClaims,
        )
        and isinstance(subject := refresh_token_claims.subject, UserId)
    ):
        user_id = subject
    if user_id:
        await token_store.log_out(user_id)
    redirect_path = "/logout" if get_env_disable_basic_auth() else "/login"
    redirect_url = prepend_root_path(request.scope, redirect_path)
    response = Response(status_code=302, headers={"Location": redirect_url})
    response = delete_access_token_cookie(response)
    response = delete_refresh_token_cookie(response)
    response = delete_oauth2_state_cookie(response)
    response = delete_oauth2_nonce_cookie(response)
    return response


async def _refresh_tokens(request: Request) -> Response:
    """Refresh access and refresh tokens."""
    if (refresh_token := request.cookies.get(PHOENIX_REFRESH_TOKEN_COOKIE_NAME)) is None:
        raise HTTPException(status_code=401, detail="Missing refresh token")
    token_store: TokenStore = request.app.state.get_token_store()
    refresh_token_claims = await token_store.read(Token(refresh_token))
    if (
        not isinstance(refresh_token_claims, RefreshTokenClaims)
        or (refresh_token_id := refresh_token_claims.token_id) is None
        or refresh_token_claims.subject is None
        or (user_id := int(refresh_token_claims.subject)) is None
        or (expiration_time := refresh_token_claims.expiration_time) is None
    ):
        raise HTTPException(status_code=401, detail="Invalid refresh token")
    if expiration_time.timestamp() <= datetime.now(timezone.utc).timestamp():
        raise HTTPException(status_code=401, detail="Expired refresh token")
    await token_store.revoke(refresh_token_id)

    if (
        (access_token := request.cookies.get(PHOENIX_ACCESS_TOKEN_COOKIE_NAME)) is not None
        and isinstance(
            access_token_claims := await token_store.read(Token(access_token)), AccessTokenClaims
        )
        and (access_token_id := access_token_claims.token_id)
    ):
        await token_store.revoke(access_token_id)

    async with request.app.state.db() as session:
        if (
            user := await session.scalar(
                select(models.User).filter_by(id=user_id).options(joinedload(models.User.role))
            )
        ) is None:
            raise HTTPException(status_code=404, detail="User not found")

    return await _create_auth_response(request, user)


async def _initiate_password_reset(request: Request) -> Response:
    """Send password reset email to user."""
    if get_env_disable_basic_auth():
        raise HTTPException(status_code=403)
    data = await request.json()
    if not (email := data.get("email")):
        raise MISSING_EMAIL

    # Sanitize email by trimming and lowercasing
    email = sanitize_email(email)

    sender: EmailSender = request.app.state.email_sender
    if sender is None:
        raise SMTP_UNAVAILABLE
    assert isinstance(token_expiry := request.app.state.password_reset_token_expiry, timedelta)
    async with request.app.state.db() as session:
        user = await session.scalar(
            select(models.User)
            .where(func.lower(models.User.email) == email)
            .options(
                joinedload(models.User.password_reset_token).load_only(models.PasswordResetToken.id)
            )
        )
    if user is None or user.auth_method != "LOCAL":
        # Withold privileged information
        return Response(status_code=204)
    token_store: TokenStore = request.app.state.get_token_store()
    if user.password_reset_token:
        await token_store.revoke(PasswordResetTokenId(user.password_reset_token.id))
    password_reset_token_claims = PasswordResetTokenClaims(
        subject=UserId(user.id),
        issued_at=datetime.now(timezone.utc),
        expiration_time=datetime.now(timezone.utc) + token_expiry,
    )
    token, _ = await token_store.create_password_reset_token(password_reset_token_claims)
    url = urlparse(request.headers.get("referer") or get_base_url())
    path = prepend_root_path(request.scope, "/reset-password-with-token")
    query_string = urlencode(dict(token=token))
    components = (url.scheme, url.netloc, path, "", query_string, "")
    reset_url = urlunparse(components)
    await sender.send_password_reset_email(email, reset_url)
    return Response(status_code=204)


async def _reset_password(request: Request) -> Response:
    """Reset user password using a valid reset token."""
    if get_env_disable_basic_auth():
        raise HTTPException(status_code=403)
    data = await request.json()
    if not (password := data.get("password")):
        raise MISSING_PASSWORD
    token_store: TokenStore = request.app.state.get_token_store()
    if (
        not (token := data.get("token"))
        or not isinstance((claims := await token_store.read(token)), PasswordResetTokenClaims)
        or not claims.expiration_time
        or claims.expiration_time <= datetime.now(timezone.utc)
    ):
        raise INVALID_TOKEN
    assert (user_id := claims.subject)
    async with request.app.state.db() as session:
        user = await session.scalar(select(models.User).filter_by(id=int(user_id)))
    if user is None or user.auth_method != "LOCAL":
        # Withold privileged information
        return Response(status_code=204)
    validate_password_format(password)
    user.password_salt = secrets.token_bytes(DEFAULT_SECRET_LENGTH)
    loop = asyncio.get_running_loop()
    user.password_hash = await loop.run_in_executor(
        None, partial(compute_password_hash, password=password, salt=user.password_salt)
    )
    user.reset_password = False
    async with request.app.state.db() as session:
        session.add(user)
        await session.flush()
    response = Response(status_code=204)
    assert (token_id := claims.token_id)
    await token_store.revoke(token_id)
    await token_store.log_out(UserId(user.id))
    return response


async def _ldap_login(request: Request) -> Response:
    """Authenticate user via LDAP and return access/refresh tokens."""
    # Use cached authenticator instance to avoid re-parsing TLS config on every request
    authenticator: LDAPAuthenticator | None = getattr(request.app.state, "ldap_authenticator", None)

    if not authenticator:
        raise HTTPException(
            status_code=503, detail="LDAP authentication is not configured on this server"
        )

    data = await request.json()
    username = data.get("username")
    password = data.get("password")

    if not username or not password:
        raise HTTPException(status_code=401, detail="Username and password required")

    # Authenticate against LDAP (reused authenticator, already parsed TLS config)
    user_info = await authenticator.authenticate(username, password)

    if not user_info:
        # Generic error message to prevent username enumeration
        raise HTTPException(status_code=401, detail="Invalid username and/or password")

    # Get or create user in Phoenix database
    async with request.app.state.db() as session:
        user = await get_or_create_ldap_user(session, user_info, authenticator.config)

    return await _create_auth_response(request, user)


async def _create_auth_response(request: Request, user: models.User) -> Response:
    """
    Creates access and refresh tokens for the user and sets them as cookies in the response.
    """
    token_store: TokenStore = request.app.state.get_token_store()
    assert isinstance(access_token_expiry := request.app.state.access_token_expiry, timedelta)
    assert isinstance(refresh_token_expiry := request.app.state.refresh_token_expiry, timedelta)

    access_token, refresh_token = await create_access_and_refresh_tokens(
        token_store=token_store,
        user=user,
        access_token_expiry=access_token_expiry,
        refresh_token_expiry=refresh_token_expiry,
    )
    response = Response(status_code=204)
    response = set_access_token_cookie(
        response=response, access_token=access_token, max_age=access_token_expiry
    )
    response = set_refresh_token_cookie(
        response=response, refresh_token=refresh_token, max_age=refresh_token_expiry
    )
    return response


LOGIN_FAILED_MESSAGE = "Invalid email and/or password"

MISSING_EMAIL = HTTPException(
    status_code=422,
    detail="Email required",
)
MISSING_PASSWORD = HTTPException(
    status_code=422,
    detail="Password required",
)
SMTP_UNAVAILABLE = HTTPException(
    status_code=503,
    detail="SMTP server not configured",
)
INVALID_TOKEN = HTTPException(
    status_code=401,
    detail="Invalid token",
)



from typing import TYPE_CHECKING, Any, Literal, Union

from typing_extensions import NotRequired, TypedDict, assert_never

if TYPE_CHECKING:
    from phoenix.server.api.helpers.prompts.models import (
        PromptToolChoiceNone,
        PromptToolChoiceOneOrMore,
        PromptToolChoiceSpecificFunctionTool,
        PromptToolChoiceZeroOrMore,
    )


class GoogleFunctionCallingConfig(TypedDict, total=False):
    """
    Based on https://github.com/googleapis/python-genai/blob/97cc7e4eafbee4fa4035e7420170ab6a2c9da7fb/google/genai/types.py#L4245
    """

    mode: NotRequired[Literal["auto", "any", "none"]]
    allowed_function_names: NotRequired[list[str]]


class GoogleToolChoice(TypedDict):
    """
    Based on https://github.com/googleapis/python-genai/blob/97cc7e4eafbee4fa4035e7420170ab6a2c9da7fb/google/genai/types.py#L4341
    """

    function_calling_config: GoogleFunctionCallingConfig


class GoogleToolChoiceConversion:
    @staticmethod
    def to_google(
        obj: Union[
            "PromptToolChoiceNone",
            "PromptToolChoiceZeroOrMore",
            "PromptToolChoiceOneOrMore",
            "PromptToolChoiceSpecificFunctionTool",
        ],
    ) -> GoogleToolChoice:
        if obj.type == "none":
            return {"function_calling_config": {"mode": "none"}}
        if obj.type == "zero_or_more":
            return {"function_calling_config": {"mode": "auto"}}
        if obj.type == "one_or_more":
            return {"function_calling_config": {"mode": "any"}}
        if obj.type == "specific_function":
            return {
                "function_calling_config": {
                    "mode": "any",
                    "allowed_function_names": [obj.function_name],
                }
            }
        assert_never(obj)

    @staticmethod
    def from_google(
        obj: Any,
    ) -> Union[
        "PromptToolChoiceNone",
        "PromptToolChoiceZeroOrMore",
        "PromptToolChoiceOneOrMore",
        "PromptToolChoiceSpecificFunctionTool",
    ]:
        from google.genai.types import ToolConfig

        from phoenix.server.api.helpers.prompts.models import (
            PromptToolChoiceNone,
            PromptToolChoiceOneOrMore,
            PromptToolChoiceSpecificFunctionTool,
            PromptToolChoiceZeroOrMore,
        )

        tool_config = ToolConfig.model_validate(obj)
        if (function_calling_config := tool_config.function_calling_config) is None:
            raise ValueError("function_calling_config is required")
        # normalize mode to lowercase since Google's API is case-insensitive
        # https://github.com/googleapis/python-genai/blob/97cc7e4eafbee4fa4035e7420170ab6a2c9da7fb/google/genai/types.py#L645
        normalized_mode = (
            function_calling_config.mode.value.lower()
            if function_calling_config.mode is not None
            else None
        )
        allowed_function_names = function_calling_config.allowed_function_names

        if allowed_function_names:
            if len(allowed_function_names) != 1:
                raise ValueError("Only one allowed function name is currently supported")
            if normalized_mode != "any":
                raise ValueError("allowed function names only supported in 'any' mode")
            return PromptToolChoiceSpecificFunctionTool(
                type="specific_function",
                function_name=allowed_function_names[0],
            )

        if normalized_mode == "none":
            return PromptToolChoiceNone(type="none")
        if normalized_mode == "auto" or normalized_mode is None:
            return PromptToolChoiceZeroOrMore(type="zero_or_more")
        if normalized_mode == "any":
            return PromptToolChoiceOneOrMore(type="one_or_more")

        raise ValueError(f"Unsupported Google tool choice mode: {normalized_mode}")



"""dataset_labels

Revision ID: 58228d933c91
Revises: 699f655af132
Create Date: 2025-09-05 17:47:34.637329

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "58228d933c91"
down_revision: Union[str, None] = "699f655af132"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None

_Integer = sa.Integer().with_variant(
    sa.BigInteger(),
    "postgresql",
)


def upgrade() -> None:
    op.create_table(
        "dataset_labels",
        sa.Column("id", _Integer, primary_key=True),
        sa.Column("name", sa.String, nullable=False, unique=True),
        sa.Column("description", sa.String, nullable=True),
        sa.Column("color", sa.String, nullable=False),
        sa.Column(
            "user_id",
            _Integer,
            sa.ForeignKey("users.id", ondelete="SET NULL"),
            nullable=True,
            index=True,
        ),
    )

    op.create_table(
        "datasets_dataset_labels",
        sa.Column(
            "dataset_id",
            _Integer,
            sa.ForeignKey("datasets.id", ondelete="CASCADE"),
            nullable=False,
        ),
        sa.Column(
            "dataset_label_id",
            _Integer,
            sa.ForeignKey("dataset_labels.id", ondelete="CASCADE"),
            nullable=False,
            # index on the second element of the composite primary key
            index=True,
        ),
        sa.PrimaryKeyConstraint(
            "dataset_id",
            "dataset_label_id",
        ),
    )


def downgrade() -> None:
    op.drop_table("datasets_dataset_labels")
    op.drop_table("dataset_labels")



from enum import Enum, auto
from functools import partial
from typing import Callable, Mapping, cast

import strawberry

from phoenix.metrics import Metric
from phoenix.metrics.metrics import Cardinality, Count, Max, Mean, Min, PercentEmpty, Quantile, Sum


@strawberry.enum
class DataQualityMetric(Enum):
    cardinality = auto()
    percentEmpty = auto()
    mean = auto()
    sum = auto()
    min = auto()
    max = auto()
    count = auto()
    p01 = auto()
    p25 = auto()
    p50 = auto()
    p75 = auto()
    p99 = auto()


DATA_QUALITY_METRIC_FACTORIES: Mapping[DataQualityMetric, Callable[[], Metric]] = {
    DataQualityMetric.cardinality: cast(Callable[[], Metric], Cardinality),
    DataQualityMetric.percentEmpty: cast(Callable[[], Metric], PercentEmpty),
    DataQualityMetric.mean: cast(Callable[[], Metric], Mean),
    DataQualityMetric.sum: cast(Callable[[], Metric], Sum),
    DataQualityMetric.min: cast(Callable[[], Metric], Min),
    DataQualityMetric.max: cast(Callable[[], Metric], Max),
    DataQualityMetric.count: cast(Callable[[], Metric], Count),
    DataQualityMetric.p01: cast(Callable[[], Metric], partial(Quantile, probability=0.01)),
    DataQualityMetric.p25: cast(Callable[[], Metric], partial(Quantile, probability=0.25)),
    DataQualityMetric.p50: cast(Callable[[], Metric], partial(Quantile, probability=0.50)),
    DataQualityMetric.p75: cast(Callable[[], Metric], partial(Quantile, probability=0.75)),
    DataQualityMetric.p99: cast(Callable[[], Metric], partial(Quantile, probability=0.99)),
}



from asyncio import get_running_loop
from dataclasses import dataclass
from functools import cached_property, partial
from pathlib import Path
from typing import Any, Optional, cast

from starlette.datastructures import Secret
from starlette.requests import Request as StarletteRequest
from starlette.responses import Response as StarletteResponse
from strawberry.fastapi import BaseContext

from phoenix.auth import (
    compute_password_hash,
)
from phoenix.core.model_schema import Model
from phoenix.db import models
from phoenix.server.api.dataloaders import (
    AnnotationConfigsByProjectDataLoader,
    AnnotationSummaryDataLoader,
    AverageExperimentRepeatedRunGroupLatencyDataLoader,
    AverageExperimentRunLatencyDataLoader,
    CacheForDataLoaders,
    DatasetDatasetSplitsDataLoader,
    DatasetExampleRevisionsDataLoader,
    DatasetExamplesAndVersionsByExperimentRunDataLoader,
    DatasetExampleSpansDataLoader,
    DatasetExampleSplitsDataLoader,
    DocumentEvaluationsDataLoader,
    DocumentEvaluationSummaryDataLoader,
    DocumentRetrievalMetricsDataLoader,
    ExperimentAnnotationSummaryDataLoader,
    ExperimentDatasetSplitsDataLoader,
    ExperimentErrorRatesDataLoader,
    ExperimentRepeatedRunGroupAnnotationSummariesDataLoader,
    ExperimentRepeatedRunGroupsDataLoader,
    ExperimentRunAnnotations,
    ExperimentRunCountsDataLoader,
    ExperimentRunsByExperimentAndExampleDataLoader,
    ExperimentSequenceNumberDataLoader,
    LastUsedTimesByGenerativeModelIdDataLoader,
    LatencyMsQuantileDataLoader,
    MinStartOrMaxEndTimeDataLoader,
    NumChildSpansDataLoader,
    NumSpansPerTraceDataLoader,
    ProjectByNameDataLoader,
    ProjectIdsByTraceRetentionPolicyIdDataLoader,
    PromptVersionSequenceNumberDataLoader,
    RecordCountDataLoader,
    SessionAnnotationsBySessionDataLoader,
    SessionIODataLoader,
    SessionNumTracesDataLoader,
    SessionNumTracesWithErrorDataLoader,
    SessionTokenUsagesDataLoader,
    SessionTraceLatencyMsQuantileDataLoader,
    SpanAnnotationsDataLoader,
    SpanByIdDataLoader,
    SpanCostBySpanDataLoader,
    SpanCostDetailsBySpanCostDataLoader,
    SpanCostDetailSummaryEntriesByGenerativeModelDataLoader,
    SpanCostDetailSummaryEntriesByProjectSessionDataLoader,
    SpanCostDetailSummaryEntriesBySpanDataLoader,
    SpanCostDetailSummaryEntriesByTraceDataLoader,
    SpanCostSummaryByExperimentDataLoader,
    SpanCostSummaryByExperimentRepeatedRunGroupDataLoader,
    SpanCostSummaryByExperimentRunDataLoader,
    SpanCostSummaryByGenerativeModelDataLoader,
    SpanCostSummaryByProjectDataLoader,
    SpanCostSummaryByProjectSessionDataLoader,
    SpanCostSummaryByTraceDataLoader,
    SpanDatasetExamplesDataLoader,
    SpanDescendantsDataLoader,
    SpanProjectsDataLoader,
    TableFieldsDataLoader,
    TokenCountDataLoader,
    TokenPricesByModelDataLoader,
    TraceAnnotationsByTraceDataLoader,
    TraceByTraceIdsDataLoader,
    TraceRetentionPolicyIdByProjectIdDataLoader,
    TraceRootSpansDataLoader,
    UserRolesDataLoader,
    UsersDataLoader,
)
from phoenix.server.api.dataloaders.dataset_labels import DatasetLabelsDataLoader
from phoenix.server.bearer_auth import PhoenixUser
from phoenix.server.daemons.span_cost_calculator import SpanCostCalculator
from phoenix.server.dml_event import DmlEvent
from phoenix.server.email.types import EmailSender
from phoenix.server.types import (
    CanGetLastUpdatedAt,
    CanPutItem,
    DbSessionFactory,
    TokenStore,
    UserId,
)


@dataclass
class DataLoaders:
    annotation_configs_by_project: AnnotationConfigsByProjectDataLoader
    annotation_summaries: AnnotationSummaryDataLoader
    average_experiment_repeated_run_group_latency: (
        AverageExperimentRepeatedRunGroupLatencyDataLoader
    )
    average_experiment_run_latency: AverageExperimentRunLatencyDataLoader
    dataset_example_fields: TableFieldsDataLoader
    dataset_example_revisions: DatasetExampleRevisionsDataLoader
    dataset_example_spans: DatasetExampleSpansDataLoader
    dataset_labels: DatasetLabelsDataLoader
    dataset_label_fields: TableFieldsDataLoader
    dataset_dataset_splits: DatasetDatasetSplitsDataLoader
    dataset_examples_and_versions_by_experiment_run: (
        DatasetExamplesAndVersionsByExperimentRunDataLoader
    )
    dataset_example_splits: DatasetExampleSplitsDataLoader
    dataset_fields: TableFieldsDataLoader
    dataset_split_fields: TableFieldsDataLoader
    dataset_version_fields: TableFieldsDataLoader
    document_annotation_fields: TableFieldsDataLoader
    document_evaluation_summaries: DocumentEvaluationSummaryDataLoader
    document_evaluations: DocumentEvaluationsDataLoader
    document_retrieval_metrics: DocumentRetrievalMetricsDataLoader
    experiment_annotation_summaries: ExperimentAnnotationSummaryDataLoader
    experiment_dataset_splits: ExperimentDatasetSplitsDataLoader
    experiment_error_rates: ExperimentErrorRatesDataLoader
    experiment_fields: TableFieldsDataLoader
    experiment_repeated_run_group_annotation_summaries: (
        ExperimentRepeatedRunGroupAnnotationSummariesDataLoader
    )
    experiment_repeated_run_groups: ExperimentRepeatedRunGroupsDataLoader
    experiment_run_annotation_fields: TableFieldsDataLoader
    experiment_run_annotations: ExperimentRunAnnotations
    experiment_run_counts: ExperimentRunCountsDataLoader
    experiment_run_fields: TableFieldsDataLoader
    experiment_runs_by_experiment_and_example: ExperimentRunsByExperimentAndExampleDataLoader
    experiment_sequence_number: ExperimentSequenceNumberDataLoader
    generative_model_fields: TableFieldsDataLoader
    last_used_times_by_generative_model_id: LastUsedTimesByGenerativeModelIdDataLoader
    latency_ms_quantile: LatencyMsQuantileDataLoader
    min_start_or_max_end_times: MinStartOrMaxEndTimeDataLoader
    num_child_spans: NumChildSpansDataLoader
    num_spans_per_trace: NumSpansPerTraceDataLoader
    project_by_name: ProjectByNameDataLoader
    project_fields: TableFieldsDataLoader
    project_trace_retention_policy_fields: TableFieldsDataLoader
    projects_by_trace_retention_policy_id: ProjectIdsByTraceRetentionPolicyIdDataLoader
    prompt_fields: TableFieldsDataLoader
    prompt_label_fields: TableFieldsDataLoader
    prompt_version_sequence_number: PromptVersionSequenceNumberDataLoader
    prompt_version_tag_fields: TableFieldsDataLoader
    project_session_annotation_fields: TableFieldsDataLoader
    project_session_fields: TableFieldsDataLoader
    record_counts: RecordCountDataLoader
    session_annotations_by_session: SessionAnnotationsBySessionDataLoader
    session_first_inputs: SessionIODataLoader
    session_last_outputs: SessionIODataLoader
    session_num_traces: SessionNumTracesDataLoader
    session_num_traces_with_error: SessionNumTracesWithErrorDataLoader
    session_token_usages: SessionTokenUsagesDataLoader
    session_trace_latency_ms_quantile: SessionTraceLatencyMsQuantileDataLoader
    span_annotation_fields: TableFieldsDataLoader
    span_annotations: SpanAnnotationsDataLoader
    span_by_id: SpanByIdDataLoader
    span_cost_by_span: SpanCostBySpanDataLoader
    span_cost_detail_fields: TableFieldsDataLoader
    span_cost_detail_summary_entries_by_generative_model: (
        SpanCostDetailSummaryEntriesByGenerativeModelDataLoader
    )
    span_cost_detail_summary_entries_by_project_session: (
        SpanCostDetailSummaryEntriesByProjectSessionDataLoader
    )
    span_cost_detail_summary_entries_by_span: SpanCostDetailSummaryEntriesBySpanDataLoader
    span_cost_detail_summary_entries_by_trace: SpanCostDetailSummaryEntriesByTraceDataLoader
    span_cost_details_by_span_cost: SpanCostDetailsBySpanCostDataLoader
    span_cost_fields: TableFieldsDataLoader
    span_cost_summary_by_experiment: SpanCostSummaryByExperimentDataLoader
    span_cost_summary_by_experiment_repeated_run_group: (
        SpanCostSummaryByExperimentRepeatedRunGroupDataLoader
    )
    span_cost_summary_by_experiment_run: SpanCostSummaryByExperimentRunDataLoader
    span_cost_summary_by_generative_model: SpanCostSummaryByGenerativeModelDataLoader
    span_cost_summary_by_project: SpanCostSummaryByProjectDataLoader
    span_cost_summary_by_project_session: SpanCostSummaryByProjectSessionDataLoader
    span_cost_summary_by_trace: SpanCostSummaryByTraceDataLoader
    span_dataset_examples: SpanDatasetExamplesDataLoader
    span_descendants: SpanDescendantsDataLoader
    span_fields: TableFieldsDataLoader
    span_projects: SpanProjectsDataLoader
    token_counts: TokenCountDataLoader
    token_prices_by_model: TokenPricesByModelDataLoader
    trace_annotation_fields: TableFieldsDataLoader
    trace_annotations_by_trace: TraceAnnotationsByTraceDataLoader
    trace_by_trace_ids: TraceByTraceIdsDataLoader
    trace_fields: TableFieldsDataLoader
    trace_retention_policy_id_by_project_id: TraceRetentionPolicyIdByProjectIdDataLoader
    trace_root_spans: TraceRootSpansDataLoader
    user_roles: UserRolesDataLoader
    user_api_key_fields: TableFieldsDataLoader
    user_fields: TableFieldsDataLoader
    users: UsersDataLoader


class _NoOp:
    def get(self, *args: Any, **kwargs: Any) -> Any: ...
    def put(self, *args: Any, **kwargs: Any) -> Any: ...


@dataclass
class Context(BaseContext):
    db: DbSessionFactory
    data_loaders: DataLoaders
    cache_for_dataloaders: Optional[CacheForDataLoaders]
    model: Model
    export_path: Path
    span_cost_calculator: SpanCostCalculator
    last_updated_at: CanGetLastUpdatedAt = _NoOp()
    event_queue: CanPutItem[DmlEvent] = _NoOp()
    corpus: Optional[Model] = None
    read_only: bool = False
    locked: bool = False
    auth_enabled: bool = False
    secret: Optional[Secret] = None
    token_store: Optional[TokenStore] = None
    email_sender: Optional[EmailSender] = None

    def get_secret(self) -> Secret:
        """A type-safe way to get the application secret. Throws an error if the secret is not set.

        Returns:
            str: the phoenix secret
        """
        if self.secret is None:
            raise ValueError(
                "Application secret not set."
                " Please set the PHOENIX_SECRET environment variable and re-deploy the application."
            )
        return self.secret

    def get_request(self) -> StarletteRequest:
        """
        A type-safe way to get the request object. Throws an error if the request is not set.
        """
        if not isinstance(request := self.request, StarletteRequest):
            raise ValueError("no request is set")
        return request

    def get_response(self) -> StarletteResponse:
        """
        A type-safe way to get the response object. Throws an error if the response is not set.
        """
        if (response := self.response) is None:
            raise ValueError("no response is set")
        return response

    async def is_valid_password(self, password: Secret, user: models.User) -> bool:
        return (
            (hash_ := user.password_hash) is not None
            and (salt := user.password_salt) is not None
            and hash_ == await self.hash_password(password, salt)
        )

    @staticmethod
    async def hash_password(password: Secret, salt: bytes) -> bytes:
        compute = partial(compute_password_hash, password=password, salt=salt)
        return await get_running_loop().run_in_executor(None, compute)

    async def log_out(self, user_id: int) -> None:
        assert self.token_store is not None
        await self.token_store.log_out(UserId(user_id))

    @cached_property
    def user(self) -> PhoenixUser:
        return cast(PhoenixUser, self.get_request().user)

    @cached_property
    def user_id(self) -> Optional[int]:
        try:
            return int(self.user.identity)
        except Exception:
            return None



from typing import Optional

from sqlalchemy import func, select, tuple_
from strawberry.dataloader import DataLoader
from typing_extensions import TypeAlias

from phoenix.db import models
from phoenix.server.types import DbSessionFactory

ExperimentID: TypeAlias = int
DatasetExampleID: TypeAlias = int
RunLatency: TypeAlias = float
Key: TypeAlias = tuple[ExperimentID, DatasetExampleID]
Result: TypeAlias = Optional[RunLatency]


class AverageExperimentRepeatedRunGroupLatencyDataLoader(DataLoader[Key, Result]):
    def __init__(
        self,
        db: DbSessionFactory,
    ) -> None:
        super().__init__(load_fn=self._load_fn)
        self._db = db

    async def _load_fn(self, keys: list[Key]) -> list[Result]:
        average_latency_query = (
            select(
                models.ExperimentRun.experiment_id.label("experiment_id"),
                models.ExperimentRun.dataset_example_id.label("example_id"),
                func.avg(models.ExperimentRun.latency_ms).label("average_repetition_latency_ms"),
            )
            .select_from(models.ExperimentRun)
            .where(
                tuple_(
                    models.ExperimentRun.experiment_id, models.ExperimentRun.dataset_example_id
                ).in_(set(keys))
            )
            .group_by(models.ExperimentRun.experiment_id, models.ExperimentRun.dataset_example_id)
        )
        async with self._db() as session:
            average_run_latencies_ms = {
                (experiment_id, example_id): average_run_latency_ms
                async for experiment_id, example_id, average_run_latency_ms in await session.stream(
                    average_latency_query
                )
            }
        return [
            average_run_latencies_ms.get((experiment_id, example_id))
            for experiment_id, example_id in keys
        ]



from typing import Optional

import strawberry
from strawberry.relay import GlobalID
from strawberry.scalars import JSON

from phoenix.server.api.exceptions import BadRequest
from phoenix.server.api.types.AnnotationSource import AnnotationSource
from phoenix.server.api.types.AnnotatorKind import AnnotatorKind


@strawberry.input
class CreateProjectSessionAnnotationInput:
    project_session_id: GlobalID
    name: str
    annotator_kind: AnnotatorKind = AnnotatorKind.HUMAN
    label: Optional[str] = None
    score: Optional[float] = None
    explanation: Optional[str] = None
    metadata: JSON = strawberry.field(default_factory=dict)
    source: AnnotationSource = AnnotationSource.APP
    identifier: Optional[str] = strawberry.UNSET

    def __post_init__(self) -> None:
        self.name = self.name.strip()
        if isinstance(self.label, str):
            self.label = self.label.strip()
        if not self.label:
            self.label = None
        if isinstance(self.explanation, str):
            self.explanation = self.explanation.strip()
        if not self.explanation:
            self.explanation = None
        if isinstance(self.identifier, str):
            self.identifier = self.identifier.strip()
        if self.score is None and not self.label and not self.explanation:
            raise BadRequest("At least one of score, label, or explanation must be not null/empty.")



from dataclasses import field, replace
from typing import Optional

import strawberry
from strawberry import UNSET
from typing_extensions import Annotated

from phoenix.core.model_schema import Column
from phoenix.metrics import Metric
from phoenix.metrics.mixins import UnaryOperator
from phoenix.server.api.types.DataQualityMetric import (
    DATA_QUALITY_METRIC_FACTORIES,
    DataQualityMetric,
)


@strawberry.input
class DataQualityMetricInput:
    metric: DataQualityMetric
    column_name: Annotated[
        Optional[str],
        strawberry.argument(
            description="When required by the metric, this is the name of the "
            "input column, e.g., to calculate mean, a specific column must be "
            "specified as input to be averaged over.",
        ),
    ] = UNSET

    metric_instance: strawberry.Private[Metric] = field(init=False)

    def __post_init__(self) -> None:
        metric_instance = DATA_QUALITY_METRIC_FACTORIES[self.metric]()
        if isinstance(metric_instance, UnaryOperator):
            if not isinstance(self.column_name, str):
                raise ValueError(f"column name must not be null for {self.metric.name}")
            metric_instance = replace(
                metric_instance,
                operand=Column(self.column_name),
            )
        self.metric_instance = metric_instance



from collections import defaultdict
from datetime import datetime
from typing import Any, Optional

import numpy as np
from aioitertools.itertools import groupby
from cachetools import LFUCache, TTLCache
from sqlalchemy import Select, select
from strawberry.dataloader import AbstractCache, DataLoader
from typing_extensions import TypeAlias

from phoenix.db import models
from phoenix.db.helpers import SupportedSQLDialect
from phoenix.metrics.retrieval_metrics import RetrievalMetrics
from phoenix.server.api.dataloaders.cache import TwoTierCache
from phoenix.server.api.input_types.TimeRange import TimeRange
from phoenix.server.api.types.DocumentEvaluationSummary import DocumentEvaluationSummary
from phoenix.server.types import DbSessionFactory
from phoenix.trace.dsl import SpanFilter

ProjectRowId: TypeAlias = int
TimeInterval: TypeAlias = tuple[Optional[datetime], Optional[datetime]]
FilterCondition: TypeAlias = Optional[str]
EvalName: TypeAlias = str

Segment: TypeAlias = tuple[ProjectRowId, TimeInterval, FilterCondition]
Param: TypeAlias = EvalName

Key: TypeAlias = tuple[ProjectRowId, Optional[TimeRange], FilterCondition, EvalName]
Result: TypeAlias = Optional[DocumentEvaluationSummary]
ResultPosition: TypeAlias = int
DEFAULT_VALUE: Result = None


def _cache_key_fn(key: Key) -> tuple[Segment, Param]:
    project_rowid, time_range, filter_condition, eval_name = key
    interval = (
        (time_range.start, time_range.end) if isinstance(time_range, TimeRange) else (None, None)
    )
    return (project_rowid, interval, filter_condition), eval_name


_Section: TypeAlias = tuple[ProjectRowId, EvalName]
_SubKey: TypeAlias = tuple[TimeInterval, FilterCondition]


class DocumentEvaluationSummaryCache(
    TwoTierCache[Key, Result, _Section, _SubKey],
):
    def __init__(self) -> None:
        super().__init__(
            # TTL=3600 (1-hour) because time intervals are always moving forward, but
            # interval endpoints are rounded down to the hour by the UI, so anything
            # older than an hour most likely won't be a cache-hit anyway.
            main_cache=TTLCache(maxsize=64 * 32, ttl=3600),
            sub_cache_factory=lambda: LFUCache(maxsize=2 * 2),
        )

    def invalidate_project(self, project_rowid: ProjectRowId) -> None:
        for section in self._cache.keys():
            if section[0] == project_rowid:
                del self._cache[section]

    def _cache_key(self, key: Key) -> tuple[_Section, _SubKey]:
        (project_rowid, interval, filter_condition), eval_name = _cache_key_fn(key)
        return (project_rowid, eval_name), (interval, filter_condition)


class DocumentEvaluationSummaryDataLoader(DataLoader[Key, Result]):
    def __init__(
        self,
        db: DbSessionFactory,
        cache_map: Optional[AbstractCache[Key, Result]] = None,
    ) -> None:
        super().__init__(
            load_fn=self._load_fn,
            cache_key_fn=_cache_key_fn,
            cache_map=cache_map,
        )
        self._db = db

    async def _load_fn(self, keys: list[Key]) -> list[Result]:
        results: list[Result] = [DEFAULT_VALUE] * len(keys)
        arguments: defaultdict[
            Segment,
            defaultdict[Param, list[ResultPosition]],
        ] = defaultdict(lambda: defaultdict(list))
        for position, key in enumerate(keys):
            segment, param = _cache_key_fn(key)
            arguments[segment][param].append(position)
        for segment, params in arguments.items():
            async with self._db() as session:
                dialect = SupportedSQLDialect(session.bind.dialect.name)
                stmt = _get_stmt(dialect, segment, *params.keys())
                data = await session.stream(stmt)
                async for eval_name, group in groupby(data, lambda d: d.name):
                    metrics_collection = []
                    async for (_, num_docs), subgroup in groupby(
                        group, lambda g: (g.id, g.num_docs)
                    ):
                        scores = [np.nan] * num_docs
                        for row in subgroup:
                            scores[row.document_position] = row.score
                        metrics_collection.append(RetrievalMetrics(scores))
                    summary = DocumentEvaluationSummary(
                        evaluation_name=eval_name,
                        metrics_collection=metrics_collection,
                    )
                    for position in params[eval_name]:
                        results[position] = summary
        return results


def _get_stmt(
    dialect: SupportedSQLDialect,
    segment: Segment,
    *eval_names: Param,
) -> Select[Any]:
    project_rowid, (start_time, end_time), filter_condition = segment
    mda = models.DocumentAnnotation
    stmt = (
        select(
            mda.name,
            models.Span.id,
            models.Span.num_documents.label("num_docs"),
            mda.score,
            mda.document_position,
        )
        .join(models.Trace)
        .where(models.Trace.project_rowid == project_rowid)
        .join(mda)
        .where(mda.name.in_(eval_names))
        .where(mda.annotator_kind == "LLM")
        .where(mda.score.is_not(None))
        .order_by(mda.name, models.Span.id)
    )
    if start_time:
        stmt = stmt.where(start_time <= models.Span.start_time)
    if end_time:
        stmt = stmt.where(models.Span.start_time < end_time)
    if filter_condition:
        span_filter = SpanFilter(condition=filter_condition)
        stmt = span_filter(stmt)
    return stmt



from typing import Optional

import strawberry


@strawberry.type
class CostBreakdown:
    tokens: Optional[float] = strawberry.field(
        default=None,
        description="Total number of tokens, including tokens for which no cost was computed.",
    )
    cost: Optional[float] = None



from enum import Enum

import strawberry


class TokenKind(Enum):
    PROMPT = "prompt"
    COMPLETION = "completion"


@strawberry.type
class TokenPrice:
    token_type: str
    kind: TokenKind
    cost_per_million_tokens: float
    cost_per_token: float



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Union

from typing_extensions import assert_never

if TYPE_CHECKING:
    from anthropic.types import (
        ToolChoiceAnyParam,
        ToolChoiceAutoParam,
        ToolChoiceParam,
        ToolChoiceToolParam,
    )

    from phoenix.server.api.helpers.prompts.models import (
        PromptToolChoiceNone,
        PromptToolChoiceOneOrMore,
        PromptToolChoiceSpecificFunctionTool,
        PromptToolChoiceZeroOrMore,
    )


class AwsToolChoiceConversion:
    @staticmethod
    def to_aws(
        obj: Union[
            PromptToolChoiceNone,
            PromptToolChoiceZeroOrMore,
            PromptToolChoiceOneOrMore,
            PromptToolChoiceSpecificFunctionTool,
        ],
        disable_parallel_tool_use: Optional[bool] = None,
    ) -> ToolChoiceParam:
        if obj.type == "zero_or_more":
            choice_auto: ToolChoiceAutoParam = {"type": "auto"}
            if disable_parallel_tool_use is not None:
                choice_auto["disable_parallel_tool_use"] = disable_parallel_tool_use
            return choice_auto
        if obj.type == "one_or_more":
            choice_any: ToolChoiceAnyParam = {"type": "any"}
            if disable_parallel_tool_use is not None:
                choice_any["disable_parallel_tool_use"] = disable_parallel_tool_use
            return choice_any
        if obj.type == "specific_function":
            choice_tool: ToolChoiceToolParam = {"type": "tool", "name": obj.function_name}
            if disable_parallel_tool_use is not None:
                choice_tool["disable_parallel_tool_use"] = disable_parallel_tool_use
            return choice_tool
        if obj.type == "none":
            return {"type": "none"}
        assert_never(obj.type)

    @staticmethod
    def from_aws(
        obj: ToolChoiceParam,
    ) -> Union[
        PromptToolChoiceNone,
        PromptToolChoiceZeroOrMore,
        PromptToolChoiceOneOrMore,
        PromptToolChoiceSpecificFunctionTool,
    ]:
        from phoenix.server.api.helpers.prompts.models import (
            PromptToolChoiceNone,
            PromptToolChoiceOneOrMore,
            PromptToolChoiceSpecificFunctionTool,
            PromptToolChoiceZeroOrMore,
        )

        if obj["type"] == "auto":
            choice_zero_or_more = PromptToolChoiceZeroOrMore(type="zero_or_more")
            return choice_zero_or_more
        if obj["type"] == "any":
            choice_one_or_more = PromptToolChoiceOneOrMore(type="one_or_more")
            return choice_one_or_more
        if obj["type"] == "tool":
            choice_function_tool = PromptToolChoiceSpecificFunctionTool(
                type="specific_function",
                function_name=obj["name"],
            )
            return choice_function_tool
        if obj["type"] == "none":
            return PromptToolChoiceNone(type="none")
        assert_never(obj)



"""drop single indices from traces, project_sessions, and experiment_runs

Revision ID: 272b66ff50f8
Revises: a20694b15f82
Create Date: 2025-08-11 20:37:46.941940

"""

from typing import Sequence, Union

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "272b66ff50f8"
down_revision: Union[str, None] = "a20694b15f82"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.drop_index(
        "ix_traces_project_rowid",
        table_name="traces",
        if_exists=True,
    )
    op.drop_index(
        "ix_traces_start_time",
        table_name="traces",
        if_exists=True,
    )

    op.drop_index(
        "ix_project_sessions_project_id",
        table_name="project_sessions",
        if_exists=True,
    )
    op.drop_index(
        "ix_project_sessions_start_time",
        table_name="project_sessions",
        if_exists=True,
    )

    op.drop_index(
        "ix_experiment_runs_experiment_id",
        table_name="experiment_runs",
        if_exists=True,
    )
    op.drop_index(
        "ix_experiment_run_annotations_experiment_run_id",
        table_name="experiment_run_annotations",
        if_exists=True,
    )

    op.drop_index(
        "ix_dataset_example_revisions_dataset_example_id",
        table_name="dataset_example_revisions",
        if_exists=True,
    )

    op.drop_index(
        "ix_span_cost_details_span_cost_id",
        table_name="span_cost_details",
        if_exists=True,
    )


def downgrade() -> None:
    op.create_index(
        "ix_traces_project_rowid",
        "traces",
        ["project_rowid"],
        if_not_exists=True,
    )
    op.create_index(
        "ix_traces_start_time",
        "traces",
        ["start_time"],
        if_not_exists=True,
    )

    op.create_index(
        "ix_project_sessions_project_id",
        "project_sessions",
        ["project_id"],
        if_not_exists=True,
    )
    op.create_index(
        "ix_project_sessions_start_time",
        "project_sessions",
        ["start_time"],
        if_not_exists=True,
    )

    op.create_index(
        "ix_experiment_runs_experiment_id",
        "experiment_runs",
        ["experiment_id"],
        if_not_exists=True,
    )
    op.create_index(
        "ix_experiment_run_annotations_experiment_run_id",
        "experiment_run_annotations",
        ["experiment_run_id"],
        if_not_exists=True,
    )

    op.create_index(
        "ix_dataset_example_revisions_dataset_example_id",
        "dataset_example_revisions",
        ["dataset_example_id"],
        if_not_exists=True,
    )

    op.create_index(
        "ix_span_cost_details_span_cost_id",
        "span_cost_details",
        ["span_cost_id"],
        if_not_exists=True,
    )



from typing import (
    Optional,
)

from starlette.requests import Request
from strawberry import Info

from phoenix.server.api.context import Context
from phoenix.server.bearer_auth import PhoenixUser


def get_user(info: Info[Context, None]) -> Optional[int]:
    user_id: Optional[int] = None
    try:
        assert isinstance(request := info.context.request, Request)

        if "user" in request.scope and isinstance((user := info.context.user), PhoenixUser):
            user_id = int(user.identity)
    except AssertionError:
        # Request is not available, try to obtain user identify
        # this will also throw an assertion error if auth is not available
        # the finally block will continue execution returning None
        if info.context.user.is_authenticated:
            user_id = int(info.context.user.identity)
    finally:
        return user_id
